"""Custom build script for hatch backend"""

import glob
import os
import subprocess

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomHook(BuildHookInterface):
    def initialize(self, version, build_data):
        if self.target_name not in ["wheel", "sdist"]:
            return
        cmd = "build:prod" if version == "standard" else "build"
        osp = os.path
        here = osp.abspath(osp.dirname(__file__))
        lab_path = osp.join(here, 'ipyparallel', 'labextension')

        if os.environ.get("IPP_DISABLE_JS") == "1":
            print("Skipping js installation")
            return

        # this tells us if labextension is built at all, not if it's up-to-date
        labextension_built = glob.glob(osp.join(lab_path, "*"))
        needs_js = True
        if not osp.isdir(osp.join(here, ".git")):
            print("Installing from a dist, not a repo")
            # not in a repo, probably installing from sdist
            # could be git-archive, though!
            # skip rebuilding js if it's already present
            if labextension_built:
                print(f"Not regenerating labextension in {lab_path}")
                needs_js = False

        if needs_js:
            subprocess.check_call(['jlpm'], cwd=here)
            subprocess.check_call(['jlpm', 'run', cmd], cwd=here)



import atexit
import sys
import time
from subprocess import Popen

import ipyparallel as ipp
from benchmarks.throughput import wait_for


def start_cluster(depth, number_of_engines, path='', log_output_to_file=False):
    ipcontroller_cmd = (
        f'{path}ipcontroller --profile=asv --nodb '
        f'--cluster-id=depth_{depth} '
        f'--HubFactory.broadcast_scheduler_depth={depth} '
        f'--HubFactory.db_class=NoDB'
    )
    print(ipcontroller_cmd)
    ipengine_cmd = f'{path}ipengine --profile=asv --cluster-id=depth_{depth} '
    ps = [
        Popen(
            ipcontroller_cmd.split(),
            stdout=(
                open('ipcontroller_output.log', 'a+')
                if log_output_to_file
                else sys.stdout
            ),
            stderr=(
                open('ipcontroller_error_output.log', 'a+')
                if log_output_to_file
                else sys.stdout
            ),
            stdin=sys.stdin,
        )
    ]
    time.sleep(2)
    client = ipp.Client(profile='asv', cluster_id=f'depth_{depth}')
    print(ipengine_cmd)
    for i in range(number_of_engines):
        ps.append(
            Popen(
                ipengine_cmd.split(),
                stdout=(
                    open('ipengine_output.log', 'a+')
                    if log_output_to_file
                    else sys.stdout
                ),
                stderr=(
                    open('ipengine_error_output.log', 'a+')
                    if log_output_to_file
                    else sys.stdout
                ),
                stdin=sys.stdin,
            )
        )
        if i % 10 == 0:
            wait_for(lambda: len(client) >= i - 10)
        if i % 20 == 0:
            time.sleep(2)
            print(f'{len(client)} engines started')

    return ps


if __name__ == '__main__':
    if len(sys.argv) > 3:
        depth = sys.argv[1]
        number_of_engines = int(sys.argv[3])
    else:
        depth = 3
        number_of_engines = 30

    ps = start_cluster(depth, number_of_engines)

    for p in ps:
        p.wait()

    def clean_up():
        for p in ps:
            p.kill()

    atexit.register(clean_up)



import shutil

import pytest

from .test_cluster import (
    test_get_output,  # noqa: F401
    test_restart_engines,  # noqa: F401
    test_signal_engines,  # noqa: F401
    test_start_stop_cluster,  # noqa: F401
    test_to_from_dict,  # noqa: F401
)

# import tests that use engine_launcher_class fixture


# override engine_launcher_class
@pytest.fixture
def engine_launcher_class():
    if shutil.which("mpiexec") is None:
        pytest.skip("Requires mpiexec")
    return 'mpi'


@pytest.fixture
def controller_launcher_class():
    if shutil.which("mpiexec") is None:
        pytest.skip("Requires mpiexec")
    return 'mpi'



import os

master_project_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), os.pardir)
)
ALL_RESULTS_DIRECTORY = os.path.join(master_project_path, 'results', 'profiling')


def get_latest_results_dir():
    return os.path.join(
        'results',
        'profiling',
        max(
            dirname
            for dirname in os.listdir(ALL_RESULTS_DIRECTORY)
            if 'initial_results' not in dirname
        ),
    )


def get_initial_results_dir():
    return os.path.join(
        'results',
        'profiling',
        next(
            (
                dirname
                for dirname in os.listdir(ALL_RESULTS_DIRECTORY)
                if 'initial_results' in dirname
            ),
            max(dirname for dirname in os.listdir(ALL_RESULTS_DIRECTORY)),
        ),
    )


if __name__ == '__main__':
    print(get_latest_results_dir())



"""Test CLI application behavior"""

import glob
import json
import os
import sys
import time
import types
from subprocess import Popen, check_call, check_output
from unittest.mock import MagicMock, create_autospec, patch

import pytest
import zmq
from ipykernel import kernelapp
from ipykernel.ipkernel import IPythonKernel
from IPython.core.profiledir import ProfileDir

import ipyparallel as ipp
import ipyparallel.engine.app


def _get_output(cmd):
    out = check_output(cmd)
    if isinstance(out, bytes):
        out = out.decode('utf8', 'replace')
    return out


@pytest.mark.parametrize(
    "submod",
    [
        "cluster",
        "engine",
        "controller",
    ],
)
def test_version(submod):
    out = _get_output([sys.executable, '-m', f'ipyparallel.{submod}', '--version'])
    assert out.strip() == ipyparallel.__version__


@pytest.mark.parametrize(
    "submod",
    [
        "cluster",
        "engine",
        "controller",
    ],
)
def test_help_all(submod):
    out = _get_output([sys.executable, '-m', f'ipyparallel.{submod}', '--help-all'])


@pytest.mark.parametrize(
    "subcommand",
    [
        "list",
        "engines",
        "start",
        "stop",
        "clean",
    ],
)
def test_ipcluster_help_all(subcommand):
    out = _get_output(
        [sys.executable, '-m', 'ipyparallel.cluster', subcommand, '--help-all']
    )


def bind_kernel(engineapp):
    app = engineapp.kernel_app
    app.shell_port = app.iopub_port = app.control_port = app.stdin_port = 0
    app._bind_socket = types.MethodType(kernelapp.IPKernelApp._bind_socket, app)
    if hasattr(kernelapp.IPKernelApp, '_try_bind_socket'):
        app._try_bind_socket = types.MethodType(
            kernelapp.IPKernelApp._try_bind_socket,
            app,
        )
    app.transport = 'tcp'
    app.ip = '127.0.0.1'
    app.init_heartbeat.return_value = None
    engineapp.bind_kernel()


def test_bind_kernel(request):
    ctx = zmq.Context.instance()
    request.addfinalizer(ctx.destroy)

    class MockIPEngineApp(ipyparallel.engine.app.IPEngine):
        kernel = None
        context = ctx

    app = MockIPEngineApp()
    app.kernel_app = MagicMock(spec=kernelapp.IPKernelApp)
    # kernelapp.IPKernelApp._instance = app.kernel_app

    app.kernel = MagicMock(spec=IPythonKernel)

    def socket_spec():
        spec = create_autospec(zmq.Socket, instance=True)
        spec.FD = 2
        return spec

    app.shell_socket = socket_spec()
    app.control_socket = socket_spec()
    app.iopub_socket = socket_spec()

    bind_kernel(app)
    assert (
        app.iopub_socket.bind_to_random_port.called
        and app.iopub_socket.bind_to_random_port.call_count == 1
    )


def ipcluster_list(*args):
    return check_output(
        [sys.executable, "-m", "ipyparallel.cluster", "list"] + list(args)
    ).decode("utf8", "replace")


def test_ipcluster_list(Cluster):
    # no clusters
    out = ipcluster_list()
    assert len(out.splitlines()) == 1
    out = ipcluster_list("-o", "json")
    assert json.loads(out) == []

    with Cluster(n=2) as rc:
        out = ipcluster_list()
        head, *rest = out.strip().splitlines()
        assert len(rest) == 1
        assert rest[0].split() == [
            "default",
            rc.cluster.cluster_id,
            "True",
            "2",
            "Local",
        ]
        cluster_list = json.loads(ipcluster_list("-o", "json"))
        assert len(cluster_list) == 1
        assert cluster_list[0]['cluster']['cluster_id'] == rc.cluster.cluster_id

    # after exit, back to no clusters
    out = ipcluster_list()
    assert len(out.splitlines()) == 1
    out = ipcluster_list("-o", "json")
    assert json.loads(out) == []


@pytest.mark.parametrize("daemonize", (False, True))
def test_ipcluster_start_stop(request, ipython_dir, daemonize):
    default_profile = ProfileDir.find_profile_dir_by_name(ipython_dir)
    default_profile_dir = default_profile.location

    # cleanup the security directory to avoid leaking files from one test to the next
    def cleanup_security():
        for f in glob.glob(os.path.join(default_profile.security_dir, "*.json")):
            print(f"Cleaning up {f}")
            try:
                os.remove(f)
            except Exception as e:
                print(f"Error removing {f}: {e}")

    request.addfinalizer(cleanup_security)

    n = 2
    start_args = ["-n", str(n)]
    if daemonize:
        start_args.append("--daemonize")
    start = Popen(
        [sys.executable, "-m", "ipyparallel.cluster", "start", "--debug"] + start_args
    )
    request.addfinalizer(start.terminate)
    if daemonize:
        # if daemonize, should exit after starting
        start.wait(30)
    else:
        # wait for file to appear
        # only need this if not daemonize
        cluster_file = ipp.Cluster(
            profile_dir=default_profile_dir, cluster_id=""
        ).cluster_file
        for i in range(100):
            if os.path.isfile(cluster_file) or start.poll() is not None:
                break
            else:
                time.sleep(0.1)
        assert os.path.isfile(cluster_file)

    # list should show a file
    out = ipcluster_list()
    assert len(out.splitlines()) == 2

    # cluster running, try to connect with default args
    cluster = ipp.Cluster.from_file(log_level=10)

    try:
        with cluster.connect_client_sync() as rc:
            rc.wait_for_engines(n=2, timeout=60)
            rc[:].apply_async(os.getpid).get(timeout=10)
    except Exception:
        print("controller output")
        print(cluster.controller.get_output())
        print("engine output")
        for engine_set in cluster.engines.values():
            print(engine_set.get_output())
        raise

    # stop with ipcluster stop
    check_call([sys.executable, "-m", "ipyparallel.cluster", "stop"])
    # start should exit when this happens
    start.wait(30)

    # and ipcluster list should return empty
    out = ipcluster_list()
    assert len(out.splitlines()) == 1

    # stop all succeeds even if there's nothing to stop
    check_call([sys.executable, "-m", "ipyparallel.cluster", "stop", "--all"])


def test_ipcluster_clean(ipython_dir):
    default_profile = ProfileDir.find_profile_dir_by_name(ipython_dir)
    default_profile_dir = default_profile.location
    log_file = os.path.join(default_profile.log_dir, "test.log")
    with open(log_file, "w") as f:
        f.write("logsssss")
    cluster_file = os.path.join(default_profile.security_dir, "cluster-abc.json")
    c = ipp.Cluster()
    with open(cluster_file, 'w') as f:
        json.dump(c.to_dict(), f)
    connection_file = os.path.join(
        default_profile.security_dir, "ipcontroller-client.json"
    )
    with open(connection_file, 'w') as f:
        f.write("{}")
    check_call([sys.executable, "-m", "ipyparallel.cluster", "clean"])

    assert not os.path.exists(log_file)
    assert not os.path.exists(cluster_file)
    assert not os.path.exists(connection_file)



#!/usr/bin/env python
"""Test the performance of the task farming system.

This script submits a set of tasks via a LoadBalancedView.  The tasks
are basically just a time.sleep(t), where t is a random number between
two limits that can be configured at the command line.  To run
the script there must first be an IPython controller and engines running::

    ipcluster start -n 16

A good test to run with 16 engines is::

    python task_profiler.py -n 128 -t 0.01 -T 1.0

This should show a speedup of 13-14x.  The limitation here is that the
overhead of a single task is about 0.001-0.01 seconds.
"""

import random
import time
from optparse import OptionParser

import ipyparallel as ipp


def main():
    parser = OptionParser()
    parser.set_defaults(n=100)
    parser.set_defaults(tmin=1e-3)
    parser.set_defaults(tmax=1)
    parser.set_defaults(profile='default')

    parser.add_option("-n", type='int', dest='n', help='the number of tasks to run')
    parser.add_option(
        "-t", type='float', dest='tmin', help='the minimum task length in seconds'
    )
    parser.add_option(
        "-T", type='float', dest='tmax', help='the maximum task length in seconds'
    )
    parser.add_option(
        "-p",
        '--profile',
        type='str',
        dest='profile',
        help="the cluster profile [default: 'default']",
    )

    (opts, args) = parser.parse_args()
    assert opts.tmax >= opts.tmin, "tmax must not be smaller than tmin"

    rc = ipp.Client()
    view = rc.load_balanced_view()
    print(view)
    rc.block = True
    nengines = len(rc.ids)

    # the jobs should take a random time within a range
    times = [
        random.random() * (opts.tmax - opts.tmin) + opts.tmin for i in range(opts.n)
    ]
    stime = sum(times)

    print(f"executing {opts.n} tasks, totalling {stime:.1f} secs on {nengines} engines")
    time.sleep(1)
    start = time.perf_counter()
    amr = view.map(time.sleep, times)
    amr.get()
    stop = time.perf_counter()

    ptime = stop - start
    scale = stime / ptime

    print(f"executed {stime:.1f} secs in {ptime:.1f} secs")
    print(f"{scale:3f}x parallel performance on {nengines} engines")
    print(f"{scale / nengines:.0%} of theoretical max")


if __name__ == '__main__':
    main()



import json
import os
import pickle
import re
from datetime import datetime
from enum import Enum
from itertools import product

from utils import seconds_to_ms

from benchmarks.constants import DEFAULT_NUMBER_OF_ENGINES

RESULTS_DIR = "results"


class BenchmarkType(Enum):
    # ECHO_MANY_ARGUMENTS = 'EchoManyArguments'
    # TIME_N_TASKS = 'Engines'
    # TIME_N_TASKS_NO_DELAY_NON_BLOCKING = 'non_blocking'
    # TIME_N_TASKS_NO_DELAY = 'NoDelay'
    __order__ = 'PUSH TIME_ASYNC DEPTH_TESTING BROADCAST'
    PUSH = 'Push'
    TIME_ASYNC = 'Async'
    DEPTH_TESTING = 'DepthTesting'
    BROADCAST = 'Broadcast'


class SchedulerType(Enum):
    __order__ = 'DIRECT_VIEW LOAD_BALANCED BROADCAST_NON_COALESCING BROADCAST_COALESCING BROADCAST'
    DIRECT_VIEW = 'DirectView'
    LOAD_BALANCED = 'LoadBalanced'
    BROADCAST_NON_COALESCING = 'NonCoalescing'
    BROADCAST_COALESCING = 'Coalescing'
    BROADCAST = 'Depth'


def get_benchmark_type(benchmark_name):
    for benchmark_type in BenchmarkType:
        if benchmark_type.value.lower() in benchmark_name.lower():
            return benchmark_type
    else:
        raise NotImplementedError(
            f"Couldn't find matching benchmarkType for {benchmark_name} "
        )


def get_scheduler_type(benchmark_name):
    for scheduler_type in SchedulerType:
        if scheduler_type.value in benchmark_name:
            return scheduler_type


class BenchmarkResult:
    def __init__(self, benchmark_results_file_name):
        with open(benchmark_results_file_name) as results_file:
            results_data = json.load(results_file)

        self.results_file_name = benchmark_results_file_name
        self.date = datetime.fromtimestamp(
            results_data["date"] // 1000
        )  # Date is in ms
        self.machine_config = results_data["params"]
        self.results_dict = {
            benchmark_name: {
                'benchmark_type': get_benchmark_type(benchmark_name),
                'scheduler_type': get_scheduler_type(benchmark_name),
                'results': [
                    Result(
                        duration_in_seconds,
                        stats_for_measurement,
                        param,
                        benchmark_name,
                        get_benchmark_type(benchmark_name),
                    )
                    for duration_in_seconds, stats_for_measurement, param in zip(
                        result_data["result"],
                        result_data["stats"],
                        (
                            product(
                                result_data["params"][0],
                                result_data["params"][1],
                                result_data['params'][2],
                            )
                            if len(result_data["params"]) > 2
                            else product(
                                result_data['params'][0], result_data['params'][1]
                            )
                        ),
                    )
                ],
            }
            for benchmark_name, result_data in results_data["results"].items()
        }


def get_number_of_engines(benchmark_name):
    match = re.search('[0-9]+', benchmark_name)
    return int(match[0]) if match else DEFAULT_NUMBER_OF_ENGINES


class Result:
    def __init__(
        self,
        duration_in_seconds,
        stats_for_measurement,
        param,
        benchmark_name,
        benchmark_type,
    ):
        if not duration_in_seconds:
            self.failed = True
            return
        self.stats_for_measurement = stats_for_measurement
        self.failed = False

        self.duration_in_ms = seconds_to_ms(duration_in_seconds)
        if (
            benchmark_type is BenchmarkType.BROADCAST
            or benchmark_type is BenchmarkType.PUSH
        ):
            self.number_of_engines = int(param[0])
            self.number_of_bytes = int(param[1])
        elif benchmark_type is BenchmarkType.TIME_ASYNC:
            self.number_of_engines = int(param[0])
            self.number_of_messages = int(param[1])
        elif benchmark_type is BenchmarkType.DEPTH_TESTING:
            self.number_of_engines = int(param[0])
            self.is_coalescing = True if param[1] == 'True' else False
            self.depth = int(param[2])
        else:
            raise NotImplementedError('BenchmarkType not found in Result constructor')


def get_benchmark_results():
    results_for_machines = {
        file_name: BenchmarkResult(
            os.path.join(os.getcwd(), RESULTS_DIR, dir_content, file_name)
        )
        for dir_content in os.listdir(RESULTS_DIR)
        if (
            os.path.isdir(os.path.join(os.getcwd(), RESULTS_DIR, dir_content))
            and "asv-testing-64-2020-04-28" in dir_content
        )
        for file_name in os.listdir(f"{RESULTS_DIR}/{dir_content}")
        if "machine" not in file_name
    }

    return results_for_machines


def get_value_dict(engines_or_cores='engines', bytes_or_tasks='tasks'):
    return {
        'Duration in ms': [],
        f'Number of {bytes_or_tasks}': [],
        f'Number of {engines_or_cores}': [],
    }


def ensure_time_n_tasks_source_structure(
    datasource, delay, number_of_cores, scheduler_type
):
    if delay not in datasource:
        datasource[delay] = {number_of_cores: {scheduler_type: get_value_dict()}}
    elif number_of_cores not in datasource[delay]:
        datasource[delay][number_of_cores] = {scheduler_type: get_value_dict()}
    elif scheduler_type not in datasource[delay][number_of_cores]:
        datasource[delay][number_of_cores][scheduler_type] = get_value_dict()


def add_time_n_tasks_source(source, benchmark, number_of_cores):
    number_of_engines = benchmark['engines']
    scheduler_type = benchmark['scheduler_type']
    for duration, tasks_num, delay in [
        (result.duration_in_ms, result.number_of_tasks, seconds_to_ms(result.delay))
        for result in benchmark['results']
        if not result.failed
    ]:
        ensure_time_n_tasks_source_structure(
            source, delay, number_of_cores, scheduler_type
        )
        dict_to_append_to = source[delay][number_of_cores][scheduler_type]
        dict_to_append_to['Duration in ms'].append(duration)
        dict_to_append_to['Number of tasks'].append(tasks_num)
        dict_to_append_to['Number of engines'].append(number_of_engines)


def add_no_delay_tasks_source(source, benchmark, number_of_cores):
    scheduler_type = benchmark['scheduler_type']
    if benchmark['benchmark_type'] not in source:
        source[benchmark['benchmark_type']] = {
            scheduler_type: get_value_dict(engines_or_cores='cores')
        }
    if scheduler_type not in source[benchmark['benchmark_type']]:
        source[benchmark['benchmark_type']][scheduler_type] = get_value_dict(
            engines_or_cores='cores'
        )
    dict_to_append_to = source[benchmark['benchmark_type']][scheduler_type]

    for duration, tasks_num in [
        (result.duration_in_ms, result.number_of_tasks)
        for result in benchmark['results']
        if not result.failed
    ]:
        dict_to_append_to['Duration in ms'].append(duration)
        dict_to_append_to['Number of tasks'].append(tasks_num)
        dict_to_append_to['Number of cores'].append(number_of_cores)


def add_to_broadcast_source(source, benchmark):
    scheduler_type = benchmark['scheduler_type'].value

    if scheduler_type not in source:
        source[scheduler_type] = get_value_dict(bytes_or_tasks='bytes')

    for duration, bytes, engines in [
        (result.duration_in_ms, result.number_of_bytes, result.number_of_engines)
        for result in benchmark['results']
        if not result.failed
    ]:
        source[scheduler_type]['Duration in ms'].append(duration)
        source[scheduler_type]['Number of bytes'].append(bytes)
        source[scheduler_type]['Number of engines'].append(engines)


def add_to_async_source(source, benchmark):
    scheduler_type = benchmark['scheduler_type'].value

    if scheduler_type not in source:
        source[scheduler_type] = get_value_dict(bytes_or_tasks='messages')

    for duration, number_of_messages, engines in [
        (result.duration_in_ms, result.number_of_messages, result.number_of_engines)
        for result in benchmark['results']
        if not result.failed
    ]:
        source[scheduler_type]['Duration in ms'].append(duration)
        source[scheduler_type]['Number of messages'].append(number_of_messages)
        source[scheduler_type]['Number of engines'].append(engines)


def add_to_depth_testing_source(source, benchmark):
    scheduler_type = benchmark['scheduler_type'].value
    if scheduler_type not in source:
        source[scheduler_type] = {
            'Duration in ms': [],
            'Number of engines': [],
            'Is coalescing': [],
            'Depth': [],
        }
    for duration, number_of_engines, is_coalescing, depth in [
        (
            result.duration_in_ms,
            result.number_of_engines,
            result.is_coalescing,
            result.depth,
        )
        for result in benchmark['results']
        if not result.failed
    ]:
        source[scheduler_type]['Duration in ms'].append(duration)
        source[scheduler_type]['Number of engines'].append(number_of_engines)
        source[scheduler_type]['Is coalescing'].append(is_coalescing)
        source[scheduler_type]['Depth'].append(depth)


def add_to_echo_many_arguments_source(source, benchmark, number_of_cores):
    view_type = 'direct_view' if benchmark['is_direct_view'] else 'load_balanced'
    if 'number_of_engines' not in source and benchmark['results']:
        source['number_of_engines'] = benchmark['results'][0].number_of_engines

    if view_type not in source:
        source[view_type] = {
            'Duration in ms': [],
            'Number of arguments': [],
            'Number of cores': [],
        }

    for duration, arguments in (
        (result.duration_in_ms, result.number_of_arguments)
        for result in benchmark['results']
        if not result.failed
    ):
        source[view_type]['Duration in ms'].append(duration)
        source[view_type]['Number of arguments'].append(arguments)
        source[view_type]['Number of cores'].append(number_of_cores)


def get_number_of_cores(machine_name):
    return int(re.findall(r'\d+', machine_name)[0])


def make_source(benchmark_type, add_to_source_f, benchmark_results=None):
    if not benchmark_results:
        benchmark_results = get_benchmark_results()
    source = {}

    for benchmark_name, benchmark in benchmark_results['results_dict'].items():
        if (
            isinstance(benchmark_type, list)
            and benchmark['benchmark_type'] in benchmark_type
        ) or benchmark['benchmark_type'] == benchmark_type:
            add_to_source_f(source, benchmark)
    return source


def get_time_n_tasks_source(benchmark_results=None):
    return make_source(
        BenchmarkType.TIME_N_TASKS, add_time_n_tasks_source, benchmark_results
    )


def get_no_delay_source(benchmark_results=None):
    return make_source(
        [
            BenchmarkType.TIME_N_TASKS_NO_DELAY,
            BenchmarkType.TIME_N_TASKS_NO_DELAY_NON_BLOCKING,
        ],
        add_no_delay_tasks_source,
        benchmark_results,
    )


def get_broadcast_source(benchmark_results=None):
    return make_source(
        BenchmarkType.BROADCAST, add_to_broadcast_source, benchmark_results
    )


def get_push_source(benchmark_results=None):
    return make_source(BenchmarkType.PUSH, add_to_broadcast_source, benchmark_results)


def get_async_source(benchmark_results):
    return make_source(BenchmarkType.TIME_ASYNC, add_to_async_source, benchmark_results)


def get_echo_many_arguments_source(benchmark_results=None):
    return make_source(
        BenchmarkType.ECHO_MANY_ARGUMENTS,
        add_to_echo_many_arguments_source,
        benchmark_results,
    )


def get_depth_testing_source(benchmark_results):
    return make_source(
        BenchmarkType.DEPTH_TESTING, add_to_depth_testing_source, benchmark_results
    )


if __name__ == "__main__":
    benchmark_results = get_benchmark_results()
    combined_results = {
        'date': benchmark_results['CoalescingBroadcast.json'].date,
        'machine_config': benchmark_results['CoalescingBroadcast.json'].machine_config,
        'results_dict': {},
    }
    for result in benchmark_results.values():
        for benchmark_name, benchmark_result in result.results_dict.items():
            combined_results['results_dict'][benchmark_name] = benchmark_result

    # source = get_broadcast_source(combined_results)
    # source = get_async_source(combined_results)
    # source = get_push_source(combined_results)
    # source = get_time_n_tasks_source(benchmark_results)
    # source = get_no_delay_source(benchmark_results)
    # source = get_echo_many_arguments_source()
    source = get_depth_testing_source(combined_results)
    with open("saved_results.pkl", "wb") as saved_results:
        pickle.dump(combined_results, saved_results)



import ipyparallel as ipp

rc = ipp.Client()
rc.block = True
view = rc[:]
view.run('communicator.py')
view.execute('com = EngineCommunicator()')

# gather the connection information into a dict
ar = view.apply_async(lambda: com.info)  # noqa: F821
peers = ar.get_dict()
# this is a dict, keyed by engine ID, of the connection info for the EngineCommunicators

# connect the engines to each other:
view.apply_sync(lambda pdict: com.connect(pdict), peers)  # noqa: F821

# now all the engines are connected, and we can communicate between them:


def broadcast(client, sender, msg_name, dest_name=None, block=None):
    """broadcast a message from one engine to all others."""
    dest_name = msg_name if dest_name is None else dest_name
    client[sender].execute(f'com.publish({msg_name})', block=None)
    targets = client.ids
    targets.remove(sender)
    return client[targets].execute(f'{dest_name}=com.consume()', block=None)


def send(client, sender, targets, msg_name, dest_name=None, block=None):
    """send a message from one to one-or-more engines."""
    dest_name = msg_name if dest_name is None else dest_name

    def _send(targets, m_name):
        msg = globals()[m_name]
        return com.send(targets, msg)  # noqa: F821

    client[sender].apply_async(_send, targets, msg_name)

    return client[targets].execute(f'{dest_name}=com.recv()', block=None)



import os

os.environ['OPENBLAS_NUM_THREADS'] = '1'
import time
import timeit

import numpy as np

import ipyparallel as ipp

delay = [0]
engines = [1, 2, 16, 64, 128, 256, 512, 1024]
byte_param = [1000, 10_000, 100_000, 1_000_000, 2_000_000]

apply_replies = {}


class ArrayNotEqual(Exception):
    pass


def wait_for(condition):
    for _ in range(750):
        if condition():
            break
        else:
            time.sleep(0.1)
    if not condition():
        raise TimeoutError('wait_for took to long to finish')


def echo(x):
    return x


def make_benchmark(benchmark_name, get_view):
    class ThroughputSuite:
        param_names = ['Number of engines', 'Number of bytes']
        timer = timeit.default_timer
        timeout = 120
        params = [engines, byte_param]

        view = None
        client = None
        reply = None

        def setup(self, number_of_engines, number_of_bytes):
            self.client = ipp.Client(profile='asv')
            self.view = get_view(self)
            self.view.targets = list(range(number_of_engines))
            wait_for(lambda: len(self.client) >= number_of_engines)

        def time_broadcast(self, engines, number_of_bytes):
            self.reply = self.view.apply_sync(
                echo, np.array([0] * number_of_bytes, dtype=np.int8)
            )

        def teardown(self, *args):
            if self.client:
                self.client.close()

    return ThroughputSuite


#
# class DirectViewBroadcast(
#     make_benchmark(
#         'DirectViewBroadcast', lambda benchmark: benchmark.client.direct_view()
#     )
# ):
#     pass
#


class CoalescingBroadcast(
    make_benchmark(
        'CoalescingBroadcast',
        lambda benchmark: benchmark.client.broadcast_view(is_coalescing=True),
    )
):
    pass


#
# class NonCoalescingBroadcast(
#     make_benchmark(
#         'NonCoalescingBroadcast',
#         lambda benchmark: benchmark.client.broadcast_view(is_coalescing=False),
#     )
# ):
#     pass


#
class DepthTestingSuite:
    param_names = ['Number of engines', 'is_coalescing', 'depth']
    timer = timeit.default_timer
    timeout = 120
    params = [engines, [True, False], [3, 4]]

    view = None
    client = None
    reply = None

    def setup(self, number_of_engines, is_coalescing, depth):
        self.client = ipp.Client(profile='asv', cluster_id=f'depth_{depth}')
        self.view = self.client.broadcast_view(is_coalescing=is_coalescing)
        self.view.targets = list(range(number_of_engines))

        wait_for(lambda: len(self.client) >= number_of_engines)

    def time_broadcast(self, number_of_engines, *args):
        self.reply = self.view.apply_sync(
            echo,
            np.array([0] * 1000, dtype=np.int8),
        )

    def teardown(self, *args):
        replies_key = tuple(args)
        if replies_key in apply_replies:
            if any(
                not np.array_equal(new_reply, stored_reply)
                for new_reply, stored_reply in zip(
                    self.reply, apply_replies[replies_key]
                )
            ):
                raise ArrayNotEqual('DepthTestingSuite', args)
        if self.client:
            self.client.close()


number_of_messages = [1, 5, 10, 20, 50, 75, 100]


def make_multiple_message_benchmark(get_view):
    class AsyncMessagesSuite:
        param_names = ['Number of engines', 'number_of_messages']
        timer = timeit.default_timer
        timeout = 180
        params = [engines, number_of_messages]

        view = None
        client = None
        reply = None

        def setup(self, number_of_engines, number_of_messages):
            self.client = ipp.Client(profile='asv')
            self.view = get_view(self)
            self.view.targets = list(range(number_of_engines))

            wait_for(lambda: len(self.client) >= number_of_engines)

        def time_async_messages(self, number_of_engines, number_of_messages):
            replies = []
            for i in range(number_of_messages):
                reply = self.view.apply_async(echo, np.array([0] * 1000, dtype=np.int8))
                replies.append(reply)
            for reply in replies:
                reply.get()

        def teardown(self, *args):
            if self.client:
                self.client.close()

    return AsyncMessagesSuite


class DirectViewAsync(
    make_multiple_message_benchmark(lambda benchmark: benchmark.client.direct_view())
):
    pass


class CoalescingAsync(
    make_multiple_message_benchmark(
        lambda benchmark: benchmark.client.broadcast_view(is_coalescing=True)
    )
):
    pass


class NonCoalescingAsync(
    make_multiple_message_benchmark(
        lambda benchmark: benchmark.client.broadcast_view(is_coalescing=False)
    )
):
    pass


def make_push_benchmark(get_view):
    class PushMessageSuite:
        param_names = ['Number of engines', 'Number of bytes']
        timer = timeit.default_timer
        timeout = 120
        params = [engines, byte_param]

        view = None
        client = None

        def setup(self, number_of_engines, number_of_bytes):
            self.client = ipp.Client(profile='asv')
            self.view = get_view(self)
            self.view.targets = list(range(number_of_engines))
            wait_for(lambda: len(self.client) >= number_of_engines)

        def time_broadcast(self, engines, number_of_bytes):
            reply = self.view.apply_sync(
                lambda x: None, np.array([0] * number_of_bytes, dtype=np.int8)
            )

        def teardown(self, *args):
            if self.client:
                self.client.close()

    return PushMessageSuite


class DirectViewPush(
    make_push_benchmark(lambda benchmark: benchmark.client.direct_view())
):
    pass


class CoalescingPush(
    make_push_benchmark(
        lambda benchmark: benchmark.client.broadcast_view(is_coalescing=True)
    )
):
    pass


class NonCoalescingPush(
    make_push_benchmark(
        lambda benchmark: benchmark.client.broadcast_view(is_coalescing=False)
    )
):
    pass



#!/usr/bin/env python
"""
A simple python program of solving a 2D wave equation in parallel.
Domain partitioning and inter-processor communication
are done by an object of class MPIRectPartitioner2D
(which is a subclass of RectPartitioner2D and uses MPI via mpi4py)

An example of running the program is (8 processors, 4x2 partition,
400x100 grid cells)::

   $ ipcluster start --engines=MPIExec -n 8 # start 8 engines with mpiexec
   $ python parallelwave-mpi.py --grid 400 100 --partition 4 2

See also parallelwave-mpi, which runs the same program, but uses MPI
(via mpi4py) for the inter-engine communication.

Authors
-------

 * Xing Cai
 * Min Ragan-Kelley

"""

import argparse
import time

from numpy import sqrt

import ipyparallel as ipp


def setup_partitioner(index, num_procs, gnum_cells, parts):
    """create a partitioner in the engine namespace"""
    global partitioner
    p = MPIRectPartitioner2D(my_id=index, num_procs=num_procs)  # noqa: F821
    p.redim(global_num_cells=gnum_cells, num_parts=parts)
    p.prepare_communication()
    # put the partitioner into the global namespace:
    partitioner = p


def setup_solver(*args, **kwargs):
    """create a WaveSolver in the engine namespace"""
    global solver
    solver = WaveSolver(*args, **kwargs)  # noqa: F821


def wave_saver(u, x, y, t):
    """save the wave log"""
    global u_hist
    global t_hist
    t_hist.append(t)
    u_hist.append(1.0 * u)


# main program:
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    paa = parser.add_argument
    paa(
        '--grid',
        '-g',
        type=int,
        nargs=2,
        default=[100, 100],
        dest='grid',
        help="Cells in the grid, e.g. --grid 100 200",
    )
    paa(
        '--partition',
        '-p',
        type=int,
        nargs=2,
        default=None,
        help="Process partition grid, e.g. --partition 4 2 for 4x2",
    )
    paa('-c', type=float, default=1.0, help="Wave speed (I think)")
    paa('-Ly', type=float, default=1.0, help="system size (in y)")
    paa('-Lx', type=float, default=1.0, help="system size (in x)")
    paa('-t', '--tstop', type=float, default=1.0, help="Time units to run")
    paa(
        '--profile',
        type=str,
        default='default',
        help="Specify the ipcluster profile for the client to connect to.",
    )
    paa(
        '--save',
        action='store_true',
        help="Add this flag to save the time/wave history during the run.",
    )
    paa(
        '--scalar',
        action='store_true',
        help="Also run with scalar interior implementation, to see vector speedup.",
    )

    ns = parser.parse_args()
    # set up arguments
    grid = ns.grid
    partition = ns.partition
    Lx = ns.Lx
    Ly = ns.Ly
    c = ns.c
    tstop = ns.tstop
    if ns.save:
        user_action = wave_saver
    else:
        user_action = None

    num_cells = 1.0 * (grid[0] - 1) * (grid[1] - 1)
    final_test = True

    # create the Client
    rc = ipp.Client(profile=ns.profile)
    num_procs = len(rc.ids)

    if partition is None:
        partition = [1, num_procs]

    assert partition[0] * partition[1] == num_procs, (
        f"can't map partition {partition} to {num_procs} engines"
    )

    view = rc[:]
    print(f"Running {grid} system on {partition} processes until {tstop:f}")

    # functions defining initial/boundary/source conditions
    def I(x, y):
        from numpy import exp

        return 1.5 * exp(-100 * ((x - 0.5) ** 2 + (y - 0.5) ** 2))

    def f(x, y, t):
        return 0.0
        # from numpy import exp,sin
        # return 10*exp(-(x - sin(100*t))**2)

    def bc(x, y, t):
        return 0.0

    # initial imports, setup rank
    view.execute(
        '\n'.join(
            [
                "from mpi4py import MPI",
                "import numpy",
                "mpi = MPI.COMM_WORLD",
                "my_id = MPI.COMM_WORLD.Get_rank()",
            ]
        ),
        block=True,
    )

    # initialize t_hist/u_hist for saving the state at each step (optional)
    view['t_hist'] = []
    view['u_hist'] = []

    # set vector/scalar implementation details
    impl = {}
    impl['ic'] = 'vectorized'
    impl['inner'] = 'scalar'
    impl['bc'] = 'vectorized'

    # execute some files so that the classes we need will be defined on the engines:
    view.run('RectPartitioner.py')
    view.run('wavesolver.py')

    # setup remote partitioner
    # note that Reference means that the argument passed to setup_partitioner will be the
    # object named 'my_id' in the engine's namespace
    view.apply_sync(
        setup_partitioner, ipp.Reference('my_id'), num_procs, grid, partition
    )
    # wait for initial communication to complete
    view.execute('mpi.barrier()')
    # setup remote solvers
    view.apply_sync(
        setup_solver,
        I,
        f,
        c,
        bc,
        Lx,
        Ly,
        partitioner=ipp.Reference('partitioner'),
        dt=0,
        implementation=impl,
    )

    # lambda for calling solver.solve:
    def _solve(*args, **kwargs):
        return solver.solve(*args, **kwargs)

    if ns.scalar:
        impl['inner'] = 'scalar'
        # run first with element-wise Python operations for each cell
        t0 = time.time()
        ar = view.apply_async(
            _solve,
            tstop,
            dt=0,
            verbose=True,
            final_test=final_test,
            user_action=user_action,
        )
        if final_test:
            # this sum is performed element-wise as results finish
            s = sum(ar)
            # the L2 norm (RMS) of the result:
            norm = sqrt(s / num_cells)
        else:
            norm = -1
        t1 = time.time()
        print(f'scalar inner-version, Wtime={t1 - t0:g}, norm={norm:g}')

    impl['inner'] = 'vectorized'
    # setup new solvers
    view.apply_sync(
        setup_solver,
        I,
        f,
        c,
        bc,
        Lx,
        Ly,
        partitioner=ipp.Reference('partitioner'),
        dt=0,
        implementation=impl,
    )
    view.execute('mpi.barrier()')

    # run again with numpy vectorized inner-implementation
    t0 = time.time()
    ar = view.apply_async(
        _solve,
        tstop,
        dt=0,
        verbose=True,
        final_test=final_test,
        user_action=user_action,
    )
    if final_test:
        # this sum is performed element-wise as results finish
        s = sum(ar)
        # the L2 norm (RMS) of the result:
        norm = sqrt(s / num_cells)
    else:
        norm = -1
    t1 = time.time()
    print(f'vector inner-version, Wtime={t1 - t0:g}, norm={norm:g}')

    # if ns.save is True, then u_hist stores the history of u as a list
    # If the partion scheme is Nx1, then u can be reconstructed via 'gather':
    if ns.save and partition[-1] == 1:
        import matplotlib.pyplot as plt

        view.execute('u_last=u_hist[-1]')
        # map mpi IDs to IPython IDs, which may not match
        ranks = view['my_id']
        targets = range(len(ranks))
        for idx in range(len(ranks)):
            targets[idx] = ranks.index(idx)
        u_last = rc[targets].gather('u_last', block=True)
        plt.pcolor(u_last)
        plt.show()



"""Example for generating an arbitrary DAG as a dependency map.

This demo uses networkx to generate the graph.

Authors
-------
* MinRK
"""

from random import randint

import networkx as nx

import ipyparallel as parallel


def randomwait():
    import time
    from random import random

    time.sleep(random())
    return time.time()


def random_dag(nodes, edges):
    """Generate a random Directed Acyclic Graph (DAG) with a given number of nodes and edges."""
    G = nx.DiGraph()
    for i in range(nodes):
        G.add_node(i)
    while edges > 0:
        a = randint(0, nodes - 1)
        b = a
        while b == a:
            b = randint(0, nodes - 1)
        G.add_edge(a, b)
        if nx.is_directed_acyclic_graph(G):
            edges -= 1
        else:
            # we closed a loop!
            G.remove_edge(a, b)
    return G


def add_children(G, parent, level, n=2):
    """Add children recursively to a binary tree."""
    if level == 0:
        return
    for i in range(n):
        child = parent + str(i)
        G.add_node(child)
        G.add_edge(parent, child)
        add_children(G, child, level - 1, n)


def make_bintree(levels):
    """Make a symmetrical binary tree with @levels"""
    G = nx.DiGraph()
    root = '0'
    G.add_node(root)
    add_children(G, root, levels, 2)
    return G


def submit_jobs(view, G, jobs):
    """Submit jobs via client where G describes the time dependencies."""
    results = {}
    for node in nx.topological_sort(G):
        with view.temp_flags(after=[results[n] for n in G.predecessors(node)]):
            results[node] = view.apply(jobs[node])
    return results


def validate_tree(G, results):
    """Validate that jobs executed after their dependencies."""
    for node in G:
        started = results[node].metadata.started
        for parent in G.predecessors(node):
            finished = results[parent].metadata.completed
            assert started > finished, f"{node} should have happened after {parent}"


def main(nodes, edges):
    """Generate a random graph, submit jobs, then validate that the
    dependency order was enforced.
    Finally, plot the graph, with time on the x-axis, and
    in-degree on the y (just for spread).  All arrows must
    point at least slightly to the right if the graph is valid.
    """
    from matplotlib import pyplot as plt
    from matplotlib.cm import gist_rainbow
    from matplotlib.dates import date2num

    print("building DAG")
    G = random_dag(nodes, edges)
    jobs = {}
    pos = {}
    colors = {}
    for node in G:
        jobs[node] = randomwait

    client = parallel.Client()
    view = client.load_balanced_view()
    print(f"submitting {nodes} tasks with {edges} dependencies")
    results = submit_jobs(view, G, jobs)
    print("waiting for results")
    client.wait_interactive()
    for node in G:
        md = results[node].metadata
        start = date2num(md.started)
        runtime = date2num(md.completed) - start
        pos[node] = (start, runtime)
        colors[node] = md.engine_id
    validate_tree(G, results)
    nx.draw(
        G,
        pos,
        node_list=list(colors.keys()),
        node_color=list(colors.values()),
        cmap=gist_rainbow,
        with_labels=False,
    )
    x, y = zip(*pos.values())
    xmin, ymin = map(min, (x, y))
    xmax, ymax = map(max, (x, y))
    xscale = xmax - xmin
    yscale = ymax - ymin
    plt.xlim(xmin - xscale * 0.1, xmax + xscale * 0.1)
    plt.ylim(ymin - yscale * 0.1, ymax + yscale * 0.1)
    return G, results


if __name__ == '__main__':
    from matplotlib import pyplot as plt

    # main(5,10)
    main(32, 96)
    plt.show()



import numpy as np

import ipyparallel as ipp


def main():
    client = ipp.Client(profile='asv')
    # client.debug = True
    direct_view = client[:]
    direct_view.targets = list(range(5))
    direct_view_result = direct_view.apply(lambda x: x * 2, 13)
    print(direct_view_result.get())
    # load_balanced_view = client.load_balanced_view()
    # result = load_balanced_view.apply(lambda x: x * 2, 13)
    # print(result.get())
    broadcast_view = client.broadcast_view(is_coalescing=True)
    broadcast_result = broadcast_view.apply_sync(
        lambda x: x, np.array([0] * 8, dtype=np.int8)
    )

    print(broadcast_result)
    print(len(broadcast_result))
    # broadcast_view2 = client.broadcast_view(is_coalescing=False)
    # broadcast_result = broadcast_view2.apply_sync(
    #     lambda x: x, np.array([0] * 8, dtype=np.int8)
    # )
    #
    # print(broadcast_result)
    # print(len(broadcast_result))


if __name__ == '__main__':
    main()



import sys

import numpy as np

import ipyparallel as ipp


def echo(delay=0):
    def inner_echo(x, **kwargs):
        import time

        if delay:
            time.sleep(delay)
        return x

    return inner_echo


def profile_many_empty_tasks(lview, n, block=True):
    lview.map(echo(0), [None] * n, block=block)


# def profile_echo_many_arguments(lview, number_of_arguments):
#     lview.map(
#         lambda x: echo_many_arguments(*x),
#         [
#                         tuple(np.empty(1, dtype=np.int8) for n in range(number_of_arguments))
#             for x in range(16)
#         ],
#         block=False,
#     )


def profile_tasks_with_large_data(lview, num_bytes):
    for _ in range(10):
        for i in range(10):
            lview.apply_sync(echo(0), np.array([0] * num_bytes, dtype=np.int8))


def run_profiling(selected_profiling_task, selected_view):
    client = ipp.Client(profile='asv')
    # add it to path on the engines
    # client[:].apply_sync(add_to_path, master_project_parent)
    print('profiling task: ', selected_profiling_task)
    print('profiling view: ', selected_view)
    if selected_view == 'direct':
        view = client[:]
    elif selected_view == 'spanning_tree':
        view = client.spanning_tree_view()
    else:
        view = client.load_balanced_view()

    # view = client[:] if selected_view == 'direct' else client.load_balanced_view()

    if selected_profiling_task == 'many_empty_tasks':
        for x in range(1, 5):
            profile_many_empty_tasks(view, 10**x)
    elif selected_profiling_task == 'many_empty_tasks_non_blocking':
        for x in range(1, 5):
            profile_many_empty_tasks(view, 10**x, block=False)
    elif selected_profiling_task == 'tasks_with_large_data':
        for x in range(1, 8):
            profile_tasks_with_large_data(view, 10**x)
    # elif selected_profiling_task == 'echo_many_arguments':
    #     for i in range(100):
    #         for number_of_arguments in ((2 ** x) - 1 for x in range(1, 9)):
    #             profile_echo_many_arguments(view, number_of_arguments)


if __name__ == "__main__":
    run_profiling(sys.argv[1], sys.argv[2])



#!/usr/bin/env python
"""
Shell helper application for OS independent shell commands

Currently, the following command are supported:
* start:   starts a process into background and returns its pid
* running: check if a process with a given pid is running
* kill:    kill a process with a given pid
* mkdir:   creates directory recursively (no error if directory already exists)
* rmdir:   remove directory recursively
* exists:  checks if a file/directory exists
* remove:  removes a file
"""

import base64
import inspect
import json
import logging
import pathlib
import re
import sys
import time
import warnings
from subprocess import CalledProcessError, Popen, TimeoutExpired, check_output
from tempfile import NamedTemporaryFile

from . import shellcmd_receive

_shell_cmd_receive_tpl = '''
import base64, json

{define_receive}

def _decode(b64_json):
    return json.loads(base64.b64decode(b64_json).decode("utf8"))

inputs_encoded = "{inputs_encoded}"
inputs = _decode(inputs_encoded)
receive_parameters=inputs["receive_parameters"]
command_parameters=inputs["command_parameters"]

with ShellCommandReceive(**receive_parameters) as r:
    r.{method}(**command_parameters)
'''

_py_detached_tpl = '''
import json
input_json = """{json_params}"""
inputs = json.loads(json_params)

from subprocess import run, PIPE, STDOUT

with open(inputs['input_filename'], 'r') as f:
    stdin = f.read()

with open(inputs['output_filename'], 'w') as f:
    run(inputs["cmd_args"], input=stdin, stdout=f, stderr=STDOUT, check=True, text=True)
'''


def _encode(inputs):
    """return encoded input parameters

    Should be shell escaping-safe
    inverse is defined in _shell_cmd_receive_tpl
    """
    return base64.b64encode(json.dumps(inputs).encode("utf8")).decode("ascii")


class ShellCommandSend:
    """
    Helper class for sending shell commands in generic and OS independent form

    The class is designed to send shell commands to an ssh connection. Nevertheless, it can be
    used for send commands to different local shell as well, which is useful for testing. Since
    the concept uses the ipyparallel package (class ipyparallel.cluster.ShellCommandReceive) for
    executing the commands, it is necessary that a valid python installation (python_path) is provided.
    By calling has_python this can be evaluated. Furthermore, get_shell_info can be called to retrieve OS and
    shell information independent of a valid python installation.

    Since some operation require different handling for different platforms/shells the class gathers
    necessary information during initialization. This is done per default at object instantiation.
    Nevertheless, it can be postponed (initialize=False) but before any member function is called, the
    initialize function has to be called.

    Beside generic check_output[...] functions (equivalent to subprocess.check_output), the class provides
    specific shell commands which have a cmd_ prefix. When adding new functions make sure that an
    equivalent is added in the ShellCommandReceive class as well.

    To start processes through an ssh connection on a windows server that stays alife after the ssh connection
    is closed, the process must be started with the breakaway creation flag set. Which works fine on 'normal'
    machine and VMs is denied on windows github runners (I guess because of security reasons). Hence, it was
    necessary to implement a work-a-round to enable CI in github. In case breakaway is not supported, a detached
    ssh connection is started (by the ShellCommandSend object) which stays open until the process to start has
    finished. see _cmd_send for further details.

    The class supports a special send receiver mode (see self.send_receiver_code), which allows command sending
    without ipyparallel being installed on the 'other side'. In this mode basically the shellcmd_receive.py file
    is also transferred, which makes development and testing much easier.
    """

    package_name = "ipyparallel.shellcmd"  # package name for send the command
    output_template = re.compile(
        r"__([a-z][a-z0-9_]+)=([a-z0-9\-\.]+)__", re.IGNORECASE
    )
    receiver_code = pathlib.Path(
        inspect.getfile(shellcmd_receive)
    ).read_text()  # get full code of receiver side
    receiver_import = (
        "from ipyparallel.cluster.shellcmd_receive import ShellCommandReceive"
    )

    def __init__(
        self,
        shell,
        args,
        python_path,
        initialize=True,
        send_receiver_code=False,
        log=None,
    ):
        self.shell = shell
        self.args = args
        self.python_path = python_path

        # shell dependent values. Those values are determined in the initialize function
        self.shell_info = None
        self._win = sys.platform.lower().startswith("win")
        self.is_powershell = None  # flag if shell is windows powershell (requires special parameter quoting)
        self.breakaway_support = None  # flag if process creation support the break_away flag (relevant for windows only; None under linux)
        self.join_params = True  # join all cmd params into a single param. does NOT work with windows cmd
        # equivalent to os.pathsep (will be changed during initialization)
        self.pathsep = "/"

        # should be activated when developing...
        self.send_receiver_code = send_receiver_code
        self.debugging = False  # if activated an output log (~/ipp_shellcmd.log) is created on receiver side
        self.log = log or logging.getLogger(__name__)

        if initialize:
            self.initialize()

    def _check_output(self, cmd, **kwargs):
        kwargs.setdefault("text", True)
        self.log.debug("check_output %s", cmd)
        return check_output(cmd, **kwargs)

    def _runs_successful(self, cmd):
        self.log.debug("Checking if %s runs successfully", cmd)
        try:
            check_output(cmd, input="")
        except CalledProcessError as e:
            return False
        return True

    @staticmethod
    def _as_list(cmd):
        if isinstance(cmd, str):
            return [cmd]
        elif isinstance(cmd, list):
            return cmd
        else:
            raise TypeError(f"Unknown command type: {cmd!r}")

    def _cmd_start_windows_no_breakaway(self, cmd_args, py_cmd):
        # if windows platform doesn't support breakaway flag (e.g. Github Runner)
        # we need to start a detached process (as work-a-round), the runs until the
        # 'remote' process has finished. But we cannot directly start the command as detached
        # process, since redirection (for retrieving the pid) doesn't work. We need a detached
        # proxy process that redirects output the to file, that can be read by current process
        # to retrieve the pid.

        assert self._win
        from subprocess import DETACHED_PROCESS

        tmp = NamedTemporaryFile(
            mode="w", delete=False
        )  # use python to generate a tempfile name
        fo_name = tmp.name
        tmp.close()
        fi_name = (
            fo_name + "_stdin.py"
        )  # use tempfile name as basis for python script input
        with open(fi_name, "w") as f:
            f.write(py_cmd)

        # simple python code that starts the actual cmd in a detached process
        inputs = dict(
            input_filename=fi_name,
            output_filename=fo_name,
            cmd_args=cmd_args,
        )
        self.log.debug("Starting detached process with inputs %s", inputs)
        input_json = json.dumps(inputs)
        py_detached = _py_detached_tpl.format(input_json=input_json)

        # now start proxy process detached
        self.log.debug("[ShellCommandSend._cmd_send] starting detached process...")
        self.log.debug("[ShellCommandSend._cmd_send] python command: \n%s", py_cmd)
        try:
            p = Popen(
                [sys.executable, '-c', py_detached],
                close_fds=True,
                creationflags=DETACHED_PROCESS,
            )
        except Exception as e:
            self.log.error(f"[ShellCommandSend._cmd_send] detached process failed: {e}")
            raise e
        self.log.debug(
            "[ShellCommandSend._cmd_send] detached process started successful. Waiting for redirected output (pid)..."
        )

        # retrieve (remote) pid from output file
        output = ""
        while True:
            with open(fo_name) as f:
                output = f.read()
                if len(output) > 0:
                    break
                if p.poll() is not None:
                    if p.returncode != 0:
                        raise CalledProcessError(p.returncode, "cmd_start")
                    else:
                        raise Exception(
                            "internal error: no pid returned, although exit code of process was 0"
                        )

            time.sleep(0.1)  # wait a 0.1s and repeat

        if not output:
            self.log.error("[ShellCommandSend._cmd_send] no output received!")
        else:
            self.log.debug("[ShellCommandSend._cmd_send] output received: %s", output)

        return output

    def _cmd_send(self, method, **kwargs):
        # in send receiver mode it is not required that the ipyparallel.cluster.shellcmd
        # exists (or is update to date) on the 'other' side of the shell. This is particular
        # useful when doing further development without copying the adapted file before each
        # test run. Furthermore, the calls are much faster.
        receive_params = {}

        # make sure that env is a dictionary with only str entries (for key and value; value can be null as well)

        if self.debugging:
            receive_params["debugging"] = True
            receive_params["log"] = '~/ipp_shellcmd.log'
        if self.breakaway_support is False:
            receive_params["use_breakaway"] = False

        py_cmd = _shell_cmd_receive_tpl.format(
            define_receive=self.receiver_code
            if self.send_receiver_code
            else self.receiver_import,
            inputs_encoded=_encode(
                {
                    "receive_parameters": receive_params,
                    "command_parameters": kwargs,
                }
            ),
            method=method,
        )
        cmd_args = self.shell + self.args + [self.python_path]
        if method == 'start' and self.breakaway_support is False:
            return self._cmd_start_windows_no_breakaway(cmd_args, py_cmd)
        else:
            return self._check_output(cmd_args, universal_newlines=True, input=py_cmd)

    def _get_pid(self, output):
        # need to extract pid value
        values = dict(self.output_template.findall(output))
        if 'remote_pid' in values:
            return int(values['remote_pid'])
        else:
            raise RuntimeError(f"Failed to get pid from output: {output}")

    def _check_for_break_away_flag(self):
        assert self._win
        assert self.python_path is not None
        py_code = "import subprocess; subprocess.Popen(['cmd.exe', '/C'], close_fds=True, creationflags=subprocess.CREATE_BREAKAWAY_FROM_JOB); print('successful')"
        cmd = (
            self.shell
            + self.args
            + [
                self.python_path,
                '-c',
                f'"{py_code}"',
            ]
        )
        try:
            # non-zero return code (in pwsh) or empty output (in cmd), if break_away test fails
            output = self._check_output(cmd).strip()
        except Exception as e:
            # just to see which exception was thrown
            warnings.warn(
                f"Break away test exception: {e!r}",
                UserWarning,
                stacklevel=4,
            )
            output = ""

        return output == "successful"

    def initialize(self):
        """initialize necessary variables by sending an echo command that works on all OS and shells"""
        if self.shell_info:
            return
        #  example outputs on
        #   windows-powershell: OS-WIN-CMD=%OS%;OS-WIN-PW=Windows_NT;OS-LINUX=;SHELL=
        #   windows-cmd       : "OS-WIN-CMD=Windows_NT;OS-WIN-PW=$env:OS;OS-LINUX=$OSTYPE;SHELL=$SHELL"
        #   ubuntu-bash       : OS-WIN-CMD=Windows_NT;OS-WIN-PW=:OS;OS-LINUX=linux-gnu;SHELL=/bin/bash
        #   macos 11          : OS-WIN-CMD=%OS%;OS-WIN-PW=:OS;OS-LINUX=darwin20;SHELL=/bin/bash
        cmd = (
            self.shell
            + self.args
            + ['echo "OS-WIN-CMD=%OS%;OS-WIN-PW=$env:OS;OS-LINUX=$OSTYPE;SHELL=$SHELL"']
        )
        timeout = 10
        try:
            output = self._check_output(
                cmd, timeout=timeout
            )  # timeout for command is set to 10s
        except CalledProcessError as e:
            raise Exception(
                "Unable to get remote shell information. Are the ssh connection data correct?"
            )
        except TimeoutExpired as e:
            raise Exception(
                f"Timeout of {timeout}s reached while retrieving remote shell information. Are the ssh connection data correct?"
            )

        entries = output.strip().strip('\\"').strip('"').split(";")
        # filter output non valid entries: contains =$ or =% or has no value assign (.....=)
        valid_entries = list(
            filter(
                lambda e: not ("=$" in e or "=%" in e or "=:" in e or e[-1] == '='),
                entries,
            )
        )
        system = shell = None

        # currently we do not check if double entries are found
        for e in valid_entries:
            key, val = e.split("=")
            if key == "OS-WIN-CMD":
                system = val
                shell = "cmd.exe"
                self.is_powershell = False
                self._win = True
                self.join_params = (
                    False  # disable joining, since it does not work for windows cmd.exe
                )
                self.pathsep = "\\"
            elif key == "OS-WIN-PW":
                system = val
                shell = "powershell.exe"
                self.is_powershell = True
                self._win = True
                self.pathsep = "\\"
            elif key == "OS-LINUX":
                system = val
                self.is_powershell = False
                self._win = False
            elif key == "SHELL":
                shell = val

        if self._win and self.python_path is not None:
            self.breakaway_support = self._check_for_break_away_flag()  # check if break away flag is available (its not in windows github runners)

        self.shell_info = (system, shell)

    def get_shell_info(self):
        """
        get shell information
        :return: (str, str): string of system and shell
        """
        assert self.shell_info  # make sure that initialize was called already
        return self.shell_info

    def has_python(self, python_path=None):
        """Check if remote python can be started ('python --version')
        :return: bool: flag if python start was found
        """
        assert self.shell_info  # make sure that initialize was called already
        if not python_path:
            python_path = self.python_path
        cmd = self.shell + self.args + [python_path, '--version']
        return self._runs_successful(cmd)

    def check_output(self, cmd, **kwargs):
        """subprocess.check_output call using the shell connection
        :param cmd: command (str or list of strs) that should be executed
        :param kwargs: additional parameters that are passed to subprocess.check_output
        :return: output of executed command"""
        assert self.shell_info  # make sure that initialize was called already
        full_cmd = self.shell + self.args + self._as_list(cmd)
        return self._check_output(full_cmd, **kwargs)

    def check_output_python_module(self, module_params, **kwargs):
        """subprocess.check_output call based on python module call
        :param module_params: python module and parameters (str or list of strs) that should be executed
        :param kwargs: additional parameters that are passed to subprocess.check_output
        :return: output of executed command
        """
        assert self.shell_info  # make sure that initialize was called already
        cmd = (
            self.shell
            + self.args
            + [self.python_path, "-m"]
            + self._as_list(module_params)
        )
        return self._check_output(cmd, **kwargs)

    def check_output_python_code(self, python_code, **kwargs):
        """subprocess.check_output call running the provided python code
        :param python_code: code that should be executed
        :param kwargs: additional parameters that are passed to subprocess.check_output
        :return: output of executed command
        """
        assert self.shell_info  # make sure that initialize was called already
        assert "input" not in kwargs  # must not be specified
        assert "universal_newlines" not in kwargs  # must not be specified
        cmd = self.shell + self.args + [self.python_path]
        return self._check_output(
            cmd, universal_newlines=True, input=python_code, **kwargs
        )

    def cmd_start(self, cmd, env=None, output_file=None):
        """starts command into background and return remote pid
        :param cmd: command (str or list of strs) that should be started
        :param env: dictionary of environment variable that should be set/unset before starting the process
        :param output_file: stdout and stderr will be redirected to the (remote) file
        :return: pid of started process
        """
        # join commands into a single parameter. otherwise
        assert self.shell_info  # make sure that initialize was called already
        return self._get_pid(
            self._cmd_send("cmd_start", cmd=cmd, env=env, output_file=output_file)
        )

    def cmd_start_python_module(self, module_params, env=None, output_file=None):
        """start python module into background and return remote pid
        :param module_params: python module and parameters (str or list of strs) that should be executed
        :param env: dictionary of environment variable that should be set before starting the process
        :param output_file: stdout and stderr will be redirected to the (remote) file
        :return: pid of started process
        """
        assert self.shell_info  # make sure that initialize was called already
        cmd = [self.python_path, "-m"] + self._as_list(module_params)
        return self._get_pid(
            self._cmd_send("cmd_start", cmd=cmd, env=env, output_file=output_file)
        )

    def cmd_start_python_code(self, python_code, env=None, output_file=None):
        """start python with provided code into background and return remote pid
        :param python_code: code that should be executed
        :param env: dictionary of environment variable that should be set before starting the process
        :param output_file: stdout and stderr will be redirected to the (remote) file
        :return: pid of started process
        """
        assert self.shell_info  # make sure that initialize was called already
        # encoding the python code as base64 stream and decoding on the other side, remove any complicated
        # quoting strategy. We do not check the length of the code, which could exceed the shell parameter
        # string limit. Since the limit is typically > 2k, little code snippets should not cause any problems.
        encoded = base64.b64encode(python_code.encode())
        py_cmd = f'import base64;exec(base64.b64decode({encoded}).decode())'
        if self._win:
            py_cmd = f'"{py_cmd}"'
        cmd = [self.python_path, "-c", py_cmd]
        return self._get_pid(
            self._cmd_send("cmd_start", cmd=cmd, env=env, output_file=output_file)
        )

    def cmd_running(self, pid):
        """check if given (remote) pid is running"""
        assert self.shell_info  # make sure that initialize was called already
        output = self._cmd_send("cmd_running", pid=pid)
        # check output
        if "__running=1__" in output:
            return True
        elif "__running=0__" in output:
            return False
        else:
            raise Exception(
                f"Unexpected output ({output}) returned from by the running shell command"
            )

    def cmd_kill(self, pid, sig=None):
        """kill (remote) process with the given pid"""
        assert self.shell_info  # make sure that initialize was called already
        if sig:
            self._cmd_send("cmd_kill", pid=pid, sig=int(sig))
        else:
            self._cmd_send("cmd_kill", pid=pid)

    def cmd_mkdir(self, path):
        """make directory recursively"""
        assert self.shell_info  # make sure that initialize was called already
        self._cmd_send("cmd_mkdir", path=path)

    def cmd_rmdir(self, path):
        """remove directory recursively"""
        assert self.shell_info  # make sure that initialize was called already
        self._cmd_send("cmd_rmdir", path=path)

    def cmd_exists(self, path):
        """check if file/path exists"""
        assert self.shell_info  # make sure that initialize was called already
        output = self._cmd_send("cmd_exists", path=path)
        # check output
        if "__exists=1__" in output:
            return True
        elif "__exists=0__" in output:
            return False
        else:
            raise Exception(
                f"Unexpected output ({output}) returned from by the exists shell command"
            )

    def cmd_remove(self, path):
        """delete remote file"""
        assert self.shell_info  # make sure that initialize was called already
        output = self._cmd_send("cmd_remove", path=path)



"""test shell command classes"""

import json
import os
import signal
import sys
import time
import warnings

import pytest

from ipyparallel.cluster.shellcmd import ShellCommandSend

windows_py_path = 'python'
linux_py_path = '/opt/conda/bin/python3'

senders = [
    pytest.param(
        "windows",
        ShellCommandSend(["cmd.exe"], ["/C"], sys.executable, initialize=False),
        id="cmd",
    ),
    pytest.param(
        "windows",
        ShellCommandSend(
            ["cmd.exe"],
            ["/C"],
            sys.executable,
            initialize=False,
            send_receiver_code=True,
        ),
        id="cmd-src",
    ),
    pytest.param(
        "windows",
        ShellCommandSend(
            ["powershell.exe"], ["-Command"], sys.executable, initialize=False
        ),
        id="powershell",
    ),
    pytest.param(
        "windows",
        ShellCommandSend(
            ["powershell.exe"],
            ["-Command"],
            sys.executable,
            initialize=False,
            send_receiver_code=True,
        ),
        id="powershell-src",
    ),
    pytest.param(
        "windows",
        ShellCommandSend(
            ["ssh"],
            ["-p", "2222", "ciuser@127.0.0.1"],
            windows_py_path,
            initialize=False,
        ),
        id="ssh-win",
    ),
    pytest.param(
        "windows",
        ShellCommandSend(
            ["ssh"],
            ["-p", "2222", "ciuser@127.0.0.1"],
            windows_py_path,
            initialize=False,
            send_receiver_code=True,
        ),
        id="ssh-win-src",
    ),
    pytest.param(
        "wsl",
        ShellCommandSend(
            ["bash"], ["-c"], "python3", initialize=False, send_receiver_code=True
        ),
        id="wsl",
    ),
    pytest.param(
        "posix",
        ShellCommandSend(["bash"], ["-c"], "python3", initialize=False),
        id="bash",
    ),
    pytest.param(
        "posix",
        ShellCommandSend(
            ["bash"],
            ["-c"],
            "python3",
            initialize=False,
            send_receiver_code=True,
        ),
        id="bash-src",
    ),
    pytest.param(
        "posix",
        ShellCommandSend(
            ["ssh"], ["-p", "2222", "ciuser@127.0.0.1"], linux_py_path, initialize=False
        ),
        id="ssh-linux",
    ),
    pytest.param(
        "posix",
        ShellCommandSend(
            ["ssh"],
            ["-p", "2222", "ciuser@127.0.0.1"],
            linux_py_path,
            initialize=False,
            send_receiver_code=True,
        ),
        id="ssh-linux-src",
    ),
]

sender_ids = [
    "cmd",
    "cmd_src",
    "pwsh",
    "pwsh_src",
    "ssh-win",
    "ssh-win_src",
    "wsl",
    "bash",
    "bash_src",
    "ssh-linux",
    "ssh-linux_src",
]


@pytest.fixture
def shellcmd_test_cmd():
    """returns a command that runs for 5 seconds"""
    if sys.platform.startswith("win"):
        return 'ping -n 5 127.0.0.1'.split()
    else:
        return 'ping -c 5 127.0.0.1'.split()


@pytest.mark.parametrize("platform, sender", senders)
def test_shellcmds(request, platform, sender, shellcmd_test_cmd, ssh_key):
    def read_via_shell(shell, filename):
        # small helper function to read a file via shell commands
        if shell._win:
            content = shell.check_output(f"type {filename}").strip()
        else:
            content = shell.check_output(f"cat {filename}").strip()
        content = content.replace("\r\n", "\n")  # correct line endings for windows
        return content.split("\n")

    def print_file(shell, filename):
        print(f"Content of file '{filename}':")
        lines = read_via_shell(sender, filename)
        for idx, l in enumerate(lines):
            print(f"{idx:3}:{l}")

    prefix = ""
    if sys.platform.startswith("win"):
        if platform == "wsl":
            pytest.skip("wsl deactivated")  # comment to activate wsl tests
            prefix = "/home/" + os.environ["USERNAME"] + "/"
        elif platform != "windows":
            pytest.skip("other platform")

    else:
        # posix platform
        if platform != "posix":
            pytest.skip("other platform")

    if "ssh" in sender.shell and not sys.platform.startswith("win"):
        sender.shell.extend(["-i", ssh_key])
    # start tests

    # initialize sender class
    sender.initialize()

    if sender.breakaway_support is not None and not sender.breakaway_support:
        warnings.warn(
            "Break away process creation flag is not available (known issue for Github Windows Runners)",
            UserWarning,
            stacklevel=2,
        )

    # if Platform.get() == Platform.Windows:
    #    sender.breakaway_support = False  # just for testing

    info = sender.get_shell_info()
    assert len(info) == 2 and info[0] and info[1]

    test_cmd = shellcmd_test_cmd

    python_ok = sender.has_python()
    assert python_ok is True

    test_dir = prefix + "shellcmd_test"
    test_file = "testfile.txt"

    # perform basic file/directory operations
    sender.cmd_mkdir(test_dir)
    assert sender.cmd_exists(test_dir) is True

    # create a simple text file with one line (works on all platforms)
    fullpath = test_dir + sender.pathsep + test_file
    sender.check_output(f'echo "test-line" > {fullpath}')
    assert sender.cmd_exists(fullpath) is True
    output_lines = read_via_shell(sender, fullpath)
    assert len(output_lines) == 1
    assert "test-line" in output_lines[0]

    sender.cmd_remove(fullpath)
    assert sender.cmd_exists(fullpath) is False

    sender.cmd_rmdir(test_dir)
    assert sender.cmd_exists(test_dir) is False

    # do start operation test
    redirect_output_file = prefix + "output.txt"
    pid = sender.cmd_start(test_cmd, output_file=redirect_output_file)
    assert pid > 0
    if not sender.cmd_running(pid):
        print_file(sender, redirect_output_file)
    # assert sender.cmd_running(pid) is True

    sender.cmd_kill(pid, signal.SIGTERM)

    if not sender.cmd_running(pid):
        print_file(sender, redirect_output_file)

    assert sender.cmd_running(pid) is False
    assert sender.cmd_exists(redirect_output_file) is True
    sender.cmd_remove(redirect_output_file)

    # do environment setting test
    # we have a dictionary of three environment entries, where the first one is empty
    ci_dict = {
        "ssh": "",
        "interface": "tcp://*",
        "registration": 60691,
        "control": 60692,
        "mux": 60693,
        "task": 60694,
        "iopub": 60695,
        "hb_ping": 60696,
        "hb_pong": 60697,
        "broadcast": [60698, 60699],
        "key": "169b682b-337c645951e7d47723061090",
        "curve_serverkey": "null",
        "location": "host",
        "pack": "json",
        "unpack": "json",
        "signature_scheme": "hmac-sha256",
    }
    env_dict = {
        "IPP_CLUSTER_ID": "",
        "IPP_PROFILE_DIR": r"~/.ipython/profile_ssh",
        "IPP_CONNECTION_INFO": json.dumps(ci_dict),
    }
    # we use a small python script to output provided environment setting
    python_cmd = "import os;"
    python_cmd += "print('IPP_CLUSTER_ID set =', 'IPP_CLUSTER_ID' in os.environ);"
    python_cmd += "print('IPP_PROFILE_DIR =',os.environ['IPP_PROFILE_DIR']);"
    python_cmd += "print('IPP_CONNECTION_INFO =',os.environ['IPP_CONNECTION_INFO'])"

    output_file = prefix + "stdout.txt"
    # pid = sender.cmd_start([sender.python_path, "-c", python_cmd], env=env_dict, output_file=output_file)
    pid = sender.cmd_start_python_code(
        python_cmd, env=env_dict, output_file=output_file
    )

    counter = 0
    max_counter = 5
    while sender.cmd_running(pid) or counter < max_counter:
        time.sleep(1)  # sleep for 1 second to make sure that command has finished
        counter += 1

    assert sender.cmd_running(pid) is False

    output_lines = read_via_shell(sender, output_file)
    # check lines if they correspond to original dictionary (env_dict)
    assert len(output_lines) == 3
    assert output_lines[0] == 'IPP_CLUSTER_ID set = False'
    for line in output_lines[1:]:
        key, value = line.split(" = ")
        assert key in env_dict
        if isinstance(env_dict[key], dict):  # convert to dictionary if needed
            value = eval(value)
        assert env_dict[key] == value



import time

import numpy as np

import ipyparallel as parallel

nlist = map(int, np.logspace(2, 9, 16, base=2))
nlist2 = map(int, np.logspace(2, 8, 15, base=2))
tlist = map(int, np.logspace(7, 22, 16, base=2))
nt = 16


def wait(t=0):
    import time

    time.sleep(t)


def echo(s=''):
    return s


def time_throughput(nmessages, t=0, f=wait):
    client = parallel.Client()
    view = client.load_balanced_view()
    # do one ping before starting timing
    if f is echo:
        t = np.random.random(t / 8)
    view.apply_sync(echo, '')
    client.spin()
    tic = time.time()
    for i in range(nmessages):
        view.apply(f, t)
    lap = time.time()
    client.wait()
    toc = time.time()
    return lap - tic, toc - tic


def do_runs(nlist, t=0, f=wait, trials=2, runner=time_throughput):
    A = np.zeros((len(nlist), 2))
    for i, n in enumerate(nlist):
        t1 = t2 = 0
        for _ in range(trials):
            time.sleep(0.25)
            ts = runner(n, t, f)
            t1 += ts[0]
            t2 += ts[1]
        t1 /= trials
        t2 /= trials
        A[i] = (t1, t2)
        A[i] = n / A[i]
        print(n, A[i])
    return A


def do_echo(n, tlist=[0], f=echo, trials=2, runner=time_throughput):
    A = np.zeros((len(tlist), 2))
    for i, t in enumerate(tlist):
        t1 = t2 = 0
        for _ in range(trials):
            time.sleep(0.25)
            ts = runner(n, t, f)
            t1 += ts[0]
            t2 += ts[1]
        t1 /= trials
        t2 /= trials
        A[i] = (t1, t2)
        A[i] = n / A[i]
        print(t, A[i])
    return A



import os

c = get_config()  # noqa

c.Cluster.controller_ip = '0.0.0.0'
c.Cluster.engine_launcher_class = 'SSH'

ssh_key = os.path.join(os.path.dirname(__file__), "id_rsa")
c.Cluster.controller_ip = '0.0.0.0'
c.Cluster.engine_launcher_class = 'SSH'
c.SSHEngineSetLauncher.scp_args = c.SSHLauncher.ssh_args = [
    "-o",
    "UserKnownHostsFile=/dev/null",
    "-o",
    "StrictHostKeyChecking=no",
    "-i",
    ssh_key,
]
c.SSHEngineSetLauncher.engines = {"ciuser@127.0.0.1:2222": 4}
c.SSHEngineSetLauncher.remote_python = "/opt/conda/bin/python3"
c.SSHEngineSetLauncher.remote_profile_dir = "/home/ciuser/.ipython/profile_default"
c.SSHEngineSetLauncher.engine_args = ['--debug']



"""Compute statistics on the digits of pi.

This uses precomputed digits of pi from the website
of Professor Yasumasa Kanada at the University of
Tokoyo: https://www.super-computing.org/

Currently, there are only functions to read the
.txt (non-compressed, non-binary) files, but adding
support for compression and binary files would be
straightforward.

This focuses on computing the number of times that
all 1, 2, n digits sequences occur in the digits of pi.
If the digits of pi are truly random, these frequencies
should be equal.
"""

import os
from urllib.request import urlretrieve

import numpy as np
from matplotlib import pyplot as plt

# Top-level functions


def fetch_pi_file(filename):
    """This will download a segment of pi from super-computing.org
    if the file is not already present.
    """

    ftpdir = "ftp://pi.super-computing.org/.2/pi200m/"
    if os.path.exists(filename):
        # we already have it
        return
    else:
        # download it
        urlretrieve(ftpdir + filename, filename)


def compute_one_digit_freqs(filename):
    """
    Read digits of pi from a file and compute the 1 digit frequencies.
    """
    d = txt_file_to_digits(filename)
    freqs = one_digit_freqs(d)
    return freqs


def compute_two_digit_freqs(filename):
    """
    Read digits of pi from a file and compute the 2 digit frequencies.
    """
    d = txt_file_to_digits(filename)
    freqs = two_digit_freqs(d)
    return freqs


def reduce_freqs(freqlist):
    """
    Add up a list of freq counts to get the total counts.
    """
    allfreqs = np.zeros_like(freqlist[0])
    for f in freqlist:
        allfreqs += f
    return allfreqs


def compute_n_digit_freqs(filename, n):
    """
    Read digits of pi from a file and compute the n digit frequencies.
    """
    d = txt_file_to_digits(filename)
    freqs = n_digit_freqs(d, n)
    return freqs


# Read digits from a txt file


def txt_file_to_digits(filename, the_type=str):
    """
    Yield the digits of pi read from a .txt file.
    """
    with open(filename) as f:
        for line in f.readlines():
            for c in line:
                if c != '\n' and c != ' ':
                    yield the_type(c)


# Actual counting functions


def one_digit_freqs(digits, normalize=False):
    """
    Consume digits of pi and compute 1 digit freq. counts.
    """
    freqs = np.zeros(10, dtype='i4')
    for d in digits:
        freqs[int(d)] += 1
    if normalize:
        freqs = freqs / freqs.sum()
    return freqs


def two_digit_freqs(digits, normalize=False):
    """
    Consume digits of pi and compute 2 digits freq. counts.
    """
    freqs = np.zeros(100, dtype='i4')
    last = next(digits)
    this = next(digits)
    for d in digits:
        index = int(last + this)
        freqs[index] += 1
        last = this
        this = d
    if normalize:
        freqs = freqs / freqs.sum()
    return freqs


def n_digit_freqs(digits, n, normalize=False):
    """
    Consume digits of pi and compute n digits freq. counts.

    This should only be used for 1-6 digits.
    """
    freqs = np.zeros(pow(10, n), dtype='i4')
    current = np.zeros(n, dtype=int)
    for i in range(n):
        current[i] = next(digits)
    for d in digits:
        index = int(''.join(map(str, current)))
        freqs[index] += 1
        current[0:-1] = current[1:]
        current[-1] = d
    if normalize:
        freqs = freqs / freqs.sum()
    return freqs


# Plotting functions


def plot_two_digit_freqs(f2):
    """
    Plot two digits frequency counts using matplotlib.
    """
    f2_copy = f2.copy()
    f2_copy.shape = (10, 10)
    ax = plt.matshow(f2_copy)
    plt.colorbar()
    for i in range(10):
        for j in range(10):
            plt.text(i - 0.2, j + 0.2, str(j) + str(i))
    plt.ylabel('First digit')
    plt.xlabel('Second digit')
    return ax


def plot_one_digit_freqs(f1):
    """
    Plot one digit frequency counts using matplotlib.
    """
    ax = plt.plot(f1, 'bo-')
    plt.title('Single digit counts in pi')
    plt.xlabel('Digit')
    plt.ylabel('Count')
    return ax



import os
from datetime import date

from ipyparallel_master_project.benchmarks.utils import get_time_stamp

LOGS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logs')
GCLOUD_DIR = os.path.join(LOGS_DIR, 'gcloud_output')
PROFILING_DIR = os.path.join(LOGS_DIR, 'profiling')
TODAY = str(date.today())


def get_dir(main_dir):
    target_dir = f"{main_dir}/{TODAY}"
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    return target_dir


def get_profiling_log_file_name():
    return os.path.join(get_dir(PROFILING_DIR), f'profiling_{get_time_stamp()}')


def get_gcloud_log_file_name(instance_name):
    return os.path.join(get_dir(GCLOUD_DIR), instance_name)


if __name__ == "__main__":
    print(LOGS_DIR)



import os

DEFAULT_MINICONDA_PATH = os.path.join(os.getcwd(), "miniconda3/bin/:")
env = os.environ.copy()
env["PATH"] = DEFAULT_MINICONDA_PATH + env["PATH"]

from subprocess import check_call

GITHUB_TOKEN = ""  # Token for machine user
ASV_TESTS_REPO = "github.com/tomoboy/ipyparallel_master_project.git"
IPYPARALLEL_REPO = "github.com/tomoboy/ipyparallel.git"


def cmd_run(*args, log_filename=None, error_filename=None):
    if len(args) == 1:
        args = args[0].split(" ")
    print(f'$ {" ".join(args)}')
    if not log_filename and not error_filename:
        check_call(args, env=env)
    else:
        check_call(
            args,
            env=env,
            stdout=open(log_filename, 'w'),
            stderr=open(error_filename, 'w'),
        )


if __name__ == "__main__":
    cmd_run(
        f"git clone -q https://{GITHUB_TOKEN}@{ASV_TESTS_REPO}"
    )  # Get benchmarks from repo
    print("Finished cloning benchmark repo")
    # Installing ipyparallel from the dev branch
    cmd_run(f"pip install -q git+https://{GITHUB_TOKEN}@{IPYPARALLEL_REPO}")
    print("Installed ipyparallel")
    # Create profile for ipyparallel, (should maybe be copied if we want some cusom values here)
    cmd_run("ipython profile create --parallel --profile=asv")
    cmd_run('echo 120000 > /proc/sys/kernel/threads-max')
    cmd_run('echo 600000 > /proc/sys/vm/max_map_count')
    cmd_run('echo 200000 > /proc/sys/kernel/piad_max')
    cmd_run('echo "*    hard    nproc    100000" > /etc/security/limits.d')
    cmd_run('echo "*    soft    nproc    100000" > /etc/security/limits.d')



"""test View objects"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
import base64
import platform
import sys
import time
from collections import namedtuple
from tempfile import NamedTemporaryFile

import pytest
import zmq
from IPython import get_ipython
from IPython.utils.io import capture_output

import ipyparallel as ipp
from ipyparallel import AsyncHubResult, AsyncMapResult, AsyncResult, error
from ipyparallel.util import interactive

from .clienttest import (
    ClusterTestCase,
    conditional_crash,
    raises_remote,
    skip_without,
    wait,
)

point = namedtuple("point", "x y")


@pytest.mark.usefixtures('ipython')
class TestView(ClusterTestCase):
    def setup_method(self):
        # On Win XP, wait for resource cleanup, else parallel test group fails
        if platform.system() == "Windows" and platform.win32_ver()[0] == "XP":
            # 1 sec fails. 1.5 sec seems ok. Using 2 sec for margin of safety
            time.sleep(2)
        super().setup_method()

    def test_z_crash_mux(self):
        """test graceful handling of engine death (direct)"""
        self.add_engines(1)
        self.minimum_engines(2)
        eid = self.client.ids[-1]
        view = self.client[-2:]
        view.scatter('should_crash', [False, True], flatten=True)
        ar = view.apply_async(conditional_crash, ipp.Reference("should_crash"))
        with raises_remote(error.EngineError):
            ar.get(10)
        tic = time.perf_counter()
        while eid in self.client.ids and time.perf_counter() - tic < 5:
            time.sleep(0.05)
        assert eid not in self.client.ids

    def test_push_pull(self):
        """test pushing and pulling"""
        data = dict(a=10, b=1.05, c=list(range(10)), d={'e': (1, 2), 'f': 'hi'})
        t = self.client.ids[-1]
        v = self.client[t]
        push = v.push
        pull = v.pull
        v.block = True
        nengines = len(self.client)
        push({'data': data})
        d = pull('data')
        assert d == data
        self.client[:].push({'data': data})
        d = self.client[:].pull('data', block=True)
        assert d == nengines * [data]
        ar = push({'data': data}, block=False)
        assert isinstance(ar, AsyncResult)
        r = ar.get()
        ar = self.client[:].pull('data', block=False)
        assert isinstance(ar, AsyncResult)
        r = ar.get()
        assert r == nengines * [data]
        self.client[:].push(dict(a=10, b=20))
        r = self.client[:].pull(('a', 'b'), block=True)
        assert r, nengines * [[10 == 20]]

    def test_push_pull_function(self):
        "test pushing and pulling functions"

        def testf(x):
            return 2.0 * x

        t = self.client.ids[-1]
        v = self.client[t]
        v.block = True
        push = v.push
        pull = v.pull
        execute = v.execute
        push({'testf': testf})
        r = pull('testf')
        assert r(1.0) == testf(1.0)
        execute('r = testf(10)')
        r = pull('r')
        assert r == testf(10)
        ar = self.client[:].push({'testf': testf}, block=False)
        ar.get()
        ar = self.client[:].pull('testf', block=False)
        rlist = ar.get()
        for r in rlist:
            assert r(1.0) == testf(1.0)
        execute("def g(x): return x*x")
        r = pull(('testf', 'g'))
        assert (r[0](10), r[1](10)) == (testf(10), 100)

    def test_push_function_globals(self):
        """test that pushed functions have access to globals"""

        @interactive
        def geta():
            return a  # noqa: F821

        # self.add_engines(1)
        v = self.client[-1]
        v.block = True
        v['f'] = geta
        with raises_remote(NameError):
            v.execute('b=f()')
        v.execute('a=5')
        v.execute('b=f()')
        assert v['b'] == 5

    def test_push_function_defaults(self):
        """test that pushed functions preserve default args"""

        def echo(a=10):
            return a

        v = self.client[-1]
        v.block = True
        v['f'] = echo
        v.execute('b=f()')
        assert v['b'] == 10

    def test_get_result(self):
        """test getting results from the Hub."""
        c = ipp.Client(profile='iptest')
        # self.add_engines(1)
        t = c.ids[-1]
        v = c[t]
        v2 = self.client[t]
        ar = v.apply_async(wait, 1)
        # give the monitor time to notice the message
        time.sleep(0.25)
        ahr = v2.get_result(ar.msg_ids[0], owner=False)
        assert isinstance(ahr, AsyncHubResult)
        assert ahr.get() == ar.get()
        ar2 = v2.get_result(ar.msg_ids[0])
        assert not isinstance(ar2, AsyncHubResult)
        assert ahr.get() == ar2.get()
        c.close()

    def test_run_newline(self):
        """test that run appends newline to files"""
        with NamedTemporaryFile('w', delete=False) as f:
            f.write(
                """def g():
                return 5
                """
            )
        v = self.client[-1]
        v.run(f.name, block=True)
        assert v.apply_sync(lambda f: f(), ipp.Reference('g')) == 5

    def test_apply_f_kwarg(self):
        v = self.client[-1]

        def echo_kwargs(**kwargs):
            return kwargs

        kwargs = v.apply_async(echo_kwargs, f=5).get(timeout=30)
        assert kwargs == dict(f=5)

    def test_apply_tracked(self):
        """test tracking for apply"""
        # self.add_engines(1)
        t = self.client.ids[-1]
        v = self.client[t]
        v.block = False

        def echo(n=1024 * 1024, **kwargs):
            with v.temp_flags(**kwargs):
                return v.apply(lambda x: x, 'x' * n)

        ar = echo(1, track=False)
        ar.wait_for_send(5)
        assert isinstance(ar._tracker, zmq.MessageTracker)
        assert ar.sent
        ar = echo(track=True)
        ar.wait_for_send(5)
        assert isinstance(ar._tracker, zmq.MessageTracker)
        assert ar.sent == ar._tracker.done
        ar._tracker.wait()
        assert ar.sent

    def test_push_tracked(self):
        t = self.client.ids[-1]
        ns = dict(x='x' * 1024 * 1024)
        v = self.client[t]
        ar = v.push(ns, block=False, track=False)
        assert ar.sent

        ar = v.push(ns, block=False, track=True)
        ar.wait_for_send()
        assert ar.sent == ar._tracker.done
        assert ar.sent
        ar.get()

    def test_scatter_tracked(self):
        t = self.client.ids
        x = 'x' * 1024 * 1024
        ar = self.client[t].scatter('x', x, block=False, track=False)
        assert ar.sent

        ar = self.client[t].scatter('x', x, block=False, track=True)
        ar._sent_event.wait()
        assert isinstance(ar._tracker, zmq.MessageTracker)
        assert ar.sent == ar._tracker.done
        ar.wait_for_send()
        assert ar.sent
        ar.get()

    def test_remote_reference(self):
        v = self.client[-1]
        v['a'] = 123
        ra = ipp.Reference('a')
        b = v.apply_sync(lambda x: x, ra)
        assert b == 123

    def test_scatter_gather(self):
        view = self.client[:]
        seq1 = list(range(16))
        view.scatter('a', seq1)
        seq2 = view.gather('a', block=True)
        assert seq2 == seq1
        with raises_remote(NameError):
            view.gather('asdf', block=True)

    @skip_without('numpy')
    def test_scatter_gather_numpy(self):
        import numpy
        from numpy.testing import assert_array_equal

        view = self.client[:]
        a = numpy.arange(64)
        view.scatter('a', a, block=True)
        b = view.gather('a', block=True)
        assert_array_equal(b, a)

    def test_scatter_gather_lazy(self):
        """scatter/gather with targets='all'"""
        view = self.client.direct_view(targets='all')
        x = list(range(64))
        view.scatter('x', x)
        gathered = view.gather('x', block=True)
        assert gathered == x

    @skip_without('numpy')
    def test_apply_numpy(self):
        """view.apply(f, ndarray)"""
        import numpy
        from numpy.testing import assert_array_equal

        A = numpy.random.random((100, 100))
        view = self.client[-1]
        for dt in ['int32', 'uint8', 'float32', 'float64']:
            B = A.astype(dt)
            C = view.apply_sync(lambda x: x, B)
            assert_array_equal(B, C)

    @skip_without('numpy')
    def test_apply_numpy_object_dtype(self):
        """view.apply(f, ndarray) with dtype=object"""
        import numpy
        from numpy.testing import assert_array_equal

        view = self.client[-1]

        A = numpy.array([dict(a=5)])
        B = view.apply_sync(lambda x: x, A)
        assert_array_equal(A, B)

        A = numpy.array([(0, dict(b=10))], dtype=[('i', int), ('o', object)])
        B = view.apply_sync(lambda x: x, A)
        assert_array_equal(A, B)

    @skip_without('numpy')
    def test_push_pull_recarray(self):
        """push/pull recarrays"""
        import numpy
        from numpy.testing import assert_array_equal

        view = self.client[-1]

        R = numpy.array(
            [
                (1, 'hi', 0.0),
                (2**30, 'there', 2.5),
                (-99999, 'world', -12345.6789),
            ],
            [('n', int), ('s', '|S10'), ('f', float)],
        )

        view['RR'] = R
        R2 = view['RR']

        r_dtype, r_shape = view.apply_sync(
            interactive(lambda: (RR.dtype, RR.shape))  # noqa: F821
        )
        assert r_dtype == R.dtype
        assert r_shape == R.shape
        assert R2.dtype == R.dtype
        assert R2.shape == R.shape
        assert_array_equal(R2, R)

    @skip_without('pandas')
    def test_push_pull_timeseries(self):
        """push/pull pandas.Series"""
        import pandas

        ts = pandas.Series(list(range(10)))

        view = self.client[-1]

        view.push(dict(ts=ts), block=True)
        rts = view['ts']

        assert type(rts) is type(ts)
        assert (ts == rts).all()

    def test_map(self):
        view = self.client[:]

        def f(x):
            return x**2

        data = list(range(16))
        ar = view.map_async(f, data)
        assert len(ar) == len(data)
        r = ar.get()
        assert r, list(map(f == data))

    def test_map_empty_sequence(self):
        view = self.client[:]
        r = view.map_sync(lambda x: x, [])
        assert r == []

    def test_map_iterable(self):
        """test map on iterables (direct)"""
        view = self.client[:]
        # 101 is prime, so it won't be evenly distributed
        arr = range(101)
        # ensure it will be an iterator, even in Python 3
        it = iter(arr)
        r = view.map_sync(lambda x: x, it)
        assert r == list(arr)

    @skip_without('numpy')
    def test_map_numpy(self):
        """test map on numpy arrays (direct)"""
        import numpy
        from numpy.testing import assert_array_equal

        view = self.client[:]
        # 101 is prime, so it won't be evenly distributed
        arr = numpy.arange(101)
        ar = view.map_async(lambda x: x, arr)
        assert len(ar) == len(arr)
        r = ar.get()
        assert_array_equal(r, arr)

    def test_scatter_gather_nonblocking(self):
        data = list(range(16))
        view = self.client[:]
        view.scatter('a', data, block=False)
        ar = view.gather('a', block=False)
        assert ar.get() == data

    @skip_without('numpy')
    def test_scatter_gather_numpy_nonblocking(self):
        import numpy
        from numpy.testing import assert_array_equal

        a = numpy.arange(64)
        view = self.client[:]
        ar = view.scatter('a', a, block=False)
        assert isinstance(ar, AsyncResult)
        amr = view.gather('a', block=False)
        assert isinstance(amr, AsyncMapResult)
        assert_array_equal(amr.get(), a)

    def test_execute(self):
        view = self.client[:]
        # self.client.debug=True
        execute = view.execute
        ar = execute('c=30', block=False)
        assert isinstance(ar, AsyncResult)
        ar = execute('d=[0,1,2]', block=False)
        self.client.wait(ar, 1)
        assert len(ar.get()) == len(self.client)
        for c in view['c']:
            assert c == 30

    def test_abort(self):
        view = self.client[-1]
        ar = view.execute('import time; time.sleep(1)', block=False)
        ar2 = view.apply_async(lambda: 2)
        view.abort(ar2)
        ar3 = view.apply_async(lambda: 3)
        view.abort(ar3.msg_ids)
        with pytest.raises(error.TaskAborted):
            ar2.get()
        with pytest.raises(error.TaskAborted):
            ar3.get()

    def test_abort_all(self):
        """view.abort() aborts all outstanding tasks"""
        view = self.client[-1]
        ars = [view.apply_async(time.sleep, 0.25) for i in range(10)]
        view.abort()
        view.wait(timeout=5)
        for ar in ars[5:]:
            with pytest.raises(error.TaskAborted):
                ar.get()

    def test_temp_flags(self):
        view = self.client[-1]
        view.block = True
        with view.temp_flags(block=False):
            assert not view.block
        assert view.block

    def test_importer(self):
        view = self.client[-1]
        view.clear(block=True)
        with view.importer:
            import re  # noqa: F401

        @interactive
        def findall(pat, s):
            # this globals() step isn't necessary in real code
            # only to prevent a closure in the test
            re = globals()['re']  # noqa: F811
            return re.findall(pat, s)

        assert view.apply_sync(findall, r'\w+', 'hello world') == 'hello world'.split()

    def test_unicode_execute(self):
        """test executing unicode strings"""
        v = self.client[-1]
        v.block = True
        code = "a='é'"
        v.execute(code)
        assert v['a'] == 'é'

    def test_unicode_apply_result(self):
        """test unicode apply results"""
        v = self.client[-1]
        r = v.apply_sync(lambda: 'é')
        assert r == 'é'

    def test_unicode_apply_arg(self):
        """test passing unicode arguments to apply"""
        v = self.client[-1]

        @interactive
        def check_unicode(a, check):
            assert not isinstance(a, bytes), f"{a!r} is bytes, not unicode"
            assert isinstance(check, bytes), f"{check!r} is not bytes"
            assert a.encode('utf8') == check, f"{a} != {check}"

        for s in ['é', 'ßø®∫', 'asdf']:
            try:
                v.apply_sync(check_unicode, s, s.encode('utf8'))
            except error.RemoteError as e:
                if e.ename == 'AssertionError':
                    self.fail(e.evalue)
                else:
                    raise e

    def test_map_reference(self):
        """view.map(<Reference>, *seqs) should work"""
        v = self.client[:]
        v.scatter('n', self.client.ids, flatten=True)
        v.execute("f = lambda x,y: x*y")
        rf = ipp.Reference('f')
        nlist = list(range(10))
        mlist = nlist[::-1]
        expected = [m * n for m, n in zip(mlist, nlist)]
        result = v.map_sync(rf, mlist, nlist)
        assert result == expected

    def test_apply_reference(self):
        """view.apply(<Reference>, *args) should work"""
        v = self.client[:]
        v.scatter('n', self.client.ids, flatten=True)
        v.execute("f = lambda x: n*x")
        rf = ipp.Reference('f')
        result = v.apply_sync(rf, 5)
        expected = [5 * id for id in self.client.ids]
        assert result == expected

    def test_eval_reference(self):
        v = self.client[self.client.ids[0]]
        v['g'] = list(range(5))
        rg = ipp.Reference('g[0]')

        def echo(x):
            return x

        assert v.apply_sync(echo, rg) == 0

    def test_reference_nameerror(self):
        v = self.client[self.client.ids[0]]
        r = ipp.Reference('elvis_has_left')

        def echo(x):
            return x

        with raises_remote(NameError):
            v.apply_sync(echo, r)

    def test_single_engine_map(self):
        e0 = self.client[self.client.ids[0]]
        r = list(range(5))
        check = [-1 * i for i in r]
        result = e0.map_sync(lambda x: -1 * x, r)
        assert result == check

    def test_len(self):
        """len(view) makes sense"""
        e0 = self.client[self.client.ids[0]]
        assert len(e0) == 1
        v = self.client[:]
        assert len(v) == len(self.client.ids)
        v = self.client.direct_view('all')
        assert len(v) == len(self.client.ids)
        v = self.client[:2]
        assert len(v) == 2
        v = self.client[:1]
        assert len(v) == 1
        v = self.client.load_balanced_view()
        assert len(v) == len(self.client.ids)

    # begin execute tests

    def test_execute_reply(self):
        e0 = self.client[self.client.ids[0]]
        e0.block = True
        ar = e0.execute("5", silent=False)
        er = ar.get()
        assert str(er) == f"<ExecuteReply[{er.execution_count}]: 5>"
        assert er.execute_result['data']['text/plain'] == '5'

    def test_execute_reply_rich(self):
        e0 = self.client[self.client.ids[0]]
        e0.block = True
        e0.execute("from IPython.display import Image, HTML")
        ar = e0.execute("Image(data=b'garbage', format='png', width=10)", silent=False)
        er = ar.get()
        b64data = base64.b64encode(b'garbage').decode('ascii')
        data, metadata = er._repr_png_()
        assert data.strip() == b64data.strip()
        assert metadata == dict(width=10)
        ar = e0.execute("HTML('<b>bold</b>')", silent=False)
        er = ar.get()
        assert er._repr_html_() == "<b>bold</b>"

    def test_execute_reply_stdout(self):
        e0 = self.client[self.client.ids[0]]
        e0.block = True
        ar = e0.execute("print (5)", silent=False)
        er = ar.get()
        assert er.stdout.strip() == '5'

    def test_execute_result(self):
        """execute triggers execute_result with silent=False"""
        view = self.client[:]
        ar = view.execute("5", silent=False, block=True)

        expected = [{'text/plain': '5'}] * len(view)
        mimes = [out['data'] for out in ar.execute_result]
        assert mimes == expected

    def test_execute_silent(self):
        """execute does not trigger execute_result with silent=True"""
        view = self.client[:]
        ar = view.execute("5", block=True)
        expected = [None] * len(view)
        assert ar.execute_result == expected

    def test_execute_magic(self):
        """execute accepts IPython commands"""
        view = self.client[:]
        view.execute("a = 5")
        ar = view.execute("%whos", block=True)
        # this will raise, if that failed
        ar.get(5)
        for stdout in ar.stdout:
            lines = stdout.splitlines()
            assert lines[0].split(), ['Variable', 'Type' == 'Data/Info']
            found = False
            for line in lines[2:]:
                split = line.split()
                if split == ['a', 'int', '5']:
                    found = True
                    break
            assert found, f"whos output wrong: {stdout}"

    def test_execute_displaypub(self):
        """execute tracks display_pub output"""
        view = self.client[:]
        view.execute("from IPython.display import *")
        ar = view.execute("[ display(i) for i in range(5) ]", block=True)

        expected = [{'text/plain': str(j)} for j in range(5)]
        for outputs in ar.outputs:
            mimes = [out['data'] for out in outputs]
            assert mimes == expected

    def test_apply_displaypub(self):
        """apply tracks display_pub output"""
        view = self.client[:]
        view.execute("from IPython.display import *")

        @interactive
        def publish():
            [display(i) for i in range(5)]  # noqa: F821

        ar = view.apply_async(publish)
        ar.get(5)
        assert ar.wait_for_output(5)
        expected = [{'text/plain': str(j)} for j in range(5)]
        for outputs in ar.outputs:
            mimes = [out['data'] for out in outputs]
            assert mimes == expected

    def test_execute_raises(self):
        """exceptions in execute requests raise appropriately"""
        view = self.client[-1]
        ar = view.execute("1/0")
        with raises_remote(ZeroDivisionError):
            ar.get(2)

    def test_remoteerror_render_exception(self):
        """RemoteErrors get nice tracebacks"""
        view = self.client[-1]
        ar = view.execute("1/0")
        ip = get_ipython()
        ip.user_ns['ar'] = ar
        with capture_output() as io:
            ip.run_cell("ar.get(2)")

        assert 'ZeroDivisionError' in io.stdout, io.stdout

    def test_compositeerror_render_exception(self):
        """CompositeErrors get nice tracebacks"""
        view = self.client[:]
        ar = view.execute("1/0")
        ip = get_ipython()
        ip.user_ns['ar'] = ar

        with capture_output() as io:
            ip.run_cell("ar.get(2)")

        count = min(error.CompositeError.tb_limit, len(view))

        assert io.stdout.count('ZeroDivisionError'), count * 2 == io.stdout
        assert io.stdout.count('by zero'), count == io.stdout
        assert io.stdout.count(':execute'), count == io.stdout

    def test_compositeerror_truncate(self):
        """Truncate CompositeErrors with many exceptions"""
        view = self.client[:]
        requests = []
        for i in range(10):
            requests.append(view.execute("1/0"))

        ar = self.client.get_result(requests)
        try:
            ar.get()
        except error.CompositeError as _e:
            e = _e
        else:
            self.fail("Should have raised CompositeError")

        lines = e.render_traceback()
        with capture_output() as io:
            e.print_traceback()

        assert "more exceptions" in lines[-1]
        count = e.tb_limit

        assert io.stdout.count('ZeroDivisionError'), 2 * count == io.stdout
        assert io.stdout.count('by zero'), count == io.stdout
        assert io.stdout.count(':execute'), count == io.stdout

    def test_magic_pylab(self):
        """%pylab works on engines"""
        pytest.importorskip('matplotlib')
        view = self.client[-1]
        ar = view.execute("%pylab inline")
        # at least check if this raised:
        reply = ar.get(5)
        # include imports, in case user config
        ar = view.execute("plot(rand(100))", silent=False)
        reply = ar.get(5)
        assert ar.wait_for_output(5)
        assert len(reply.outputs) == 1
        output = reply.outputs[0]
        assert "data" in output
        data = output['data']
        assert "image/png" in data

    def test_func_default_func(self):
        """interactively defined function as apply func default"""

        def foo():
            return 'foo'

        def bar(f=foo):
            return f()

        view = self.client[-1]
        ar = view.apply_async(bar)
        r = ar.get(10)
        assert r == 'foo'

    def test_data_pub_single(self):
        view = self.client[-1]
        ar = view.execute(
            '\n'.join(
                [
                    'from ipyparallel.datapub import publish_data',
                    'for i in range(5):',
                    '  publish_data(dict(i=i))',
                ]
            ),
            block=False,
        )
        assert isinstance(ar.data, dict)
        ar.get(5)
        assert ar.wait_for_output(5)
        assert ar.data == dict(i=4)

    def test_data_pub(self):
        view = self.client[:]
        ar = view.execute(
            '\n'.join(
                [
                    'from ipyparallel.datapub import publish_data',
                    'for i in range(5):',
                    '  publish_data(dict(i=i))',
                ]
            ),
            block=False,
        )
        assert all(isinstance(d, dict) for d in ar.data)
        ar.get(5)
        assert ar.wait_for_output(5)
        assert ar.data == [dict(i=4)] * len(ar)

    def test_can_list_arg(self):
        """args in lists are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = ipp.Reference('a')
        ar = view.apply_async(lambda x: x, [rA])
        r = ar.get(5)
        assert r == [128]

    def test_can_dict_arg(self):
        """args in dicts are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = ipp.Reference('a')
        ar = view.apply_async(lambda x: x, dict(foo=rA))
        r = ar.get(5)
        assert r == dict(foo=128)

    def test_can_list_kwarg(self):
        """kwargs in lists are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = ipp.Reference('a')
        ar = view.apply_async(lambda x=5: x, x=[rA])
        r = ar.get(5)
        assert r == [128]

    def test_can_dict_kwarg(self):
        """kwargs in dicts are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = ipp.Reference('a')
        ar = view.apply_async(lambda x=5: x, dict(foo=rA))
        r = ar.get(5)
        assert r == dict(foo=128)

    def test_map_ref(self):
        """view.map works with references"""
        view = self.client[:]
        ranks = sorted(self.client.ids)
        view.scatter('rank', ranks, flatten=True)
        rrank = ipp.Reference('rank')

        amr = view.map_async(lambda x: x * 2, [rrank] * len(view))
        drank = amr.get(5)
        assert drank == [r * 2 for r in ranks]

    def test_nested_getitem_setitem(self):
        """get and set with view['a.b']"""
        view = self.client[-1]
        view.execute(
            '\n'.join(
                [
                    'class A: pass',
                    'a = A()',
                    'a.b = 128',
                ]
            ),
            block=True,
        )
        ra = ipp.Reference('a')

        r = view.apply_sync(lambda x: x.b, ra)
        assert r == 128
        assert view['a.b'] == 128

        view['a.b'] = 0

        r = view.apply_sync(lambda x: x.b, ra)
        assert r == 0
        assert view['a.b'] == 0

    def test_return_namedtuple(self):
        def namedtuplify(x, y):
            return point(x, y)

        view = self.client[-1]
        p = view.apply_sync(namedtuplify, 1, 2)
        assert p.x == 1
        assert p.y == 2

    def test_apply_namedtuple(self):
        def echoxy(p):
            return p.y, p.x

        view = self.client[-1]
        tup = view.apply_sync(echoxy, point(1, 2))
        assert tup, 2 == 1

    def test_sync_imports(self):
        view = self.client[-1]
        with capture_output() as io:
            with view.sync_imports():
                import IPython  # noqa
        assert "IPython" in io.stdout

        @interactive
        def find_ipython():
            return 'IPython' in globals()

        assert view.apply_sync(find_ipython)

    def test_sync_imports_quiet(self):
        view = self.client[-1]
        with capture_output() as io:
            with view.sync_imports(quiet=True):
                import IPython  # noqa
        assert io.stdout == ''

        @interactive
        def find_ipython():
            return 'IPython' in globals()

        assert view.apply_sync(find_ipython)

    @skip_without('cloudpickle')
    def test_use_cloudpickle(self):
        view = self.client[:]
        view['_a'] = 'engine'
        __main__ = sys.modules['__main__']

        @interactive
        def get_a():
            return _a  # noqa: F821

        # verify initial condition
        a_list = view.apply_sync(get_a)
        assert a_list == ['engine'] * len(view)

        # enable cloudpickle
        view.use_cloudpickle()

        # cloudpickle prefers client values
        __main__._a = 'client'
        a_list = view.apply_sync(get_a)
        assert a_list == ['client'] * len(view)

        # still works even if remote is undefined
        view.execute('del _a', block=True)
        a_list = view.apply_sync(get_a)
        assert a_list == ['client'] * len(view)

        # restore pickle, shouldn't resolve
        view.use_pickle()
        with raises_remote(NameError):
            view.apply_sync(get_a)

    @skip_without('cloudpickle')
    def test_cloudpickle_push_pull(self):
        view = self.client[:]
        # enable cloudpickle
        view.use_cloudpickle()

        # push/pull still work
        view.push({"key": "pushed"}, block=True)
        ar = view.pull("key")
        assert ar.get(timeout=10) == ["pushed"] * len(view)

        # restore pickle, should get the same value
        view.use_pickle()
        ar = view.pull("key")
        assert ar.get(timeout=10) == ["pushed"] * len(view)

    @skip_without('cloudpickle')
    @pytest.mark.xfail(reason="@require doesn't work with cloudpickle")
    def test_cloudpickle_require(self):
        view = self.client[:]
        # enable cloudpickle
        view.use_cloudpickle()
        assert 'types' not in globals(), (
            "Test condition isn't met if types is already imported"
        )

        @ipp.require("types")
        @ipp.interactive
        def func(x):
            return types.SimpleNamespace(n=x)  # noqa: F821

        res = view.apply_async(func, 5)
        assert res.get(timeout=10) == []
        view.use_pickle()

    def test_block_kwarg(self):
        """
        Tests that kwargs such as 'block' can be
        specified when creating a view, in this
        case a BroadcastView
        """
        view = self.client.broadcast_view(block=True)
        assert view.block is True
        # Execute something as a sanity check
        view.execute("5", silent=False)



import shutil
from unittest import mock

import pytest
from traitlets.config import Config

from . import test_cluster
from .conftest import temporary_ipython_dir
from .test_cluster import (
    test_get_output,  # noqa: F401
    test_restart_engines,  # noqa: F401
    test_signal_engines,  # noqa: F401
    test_start_stop_cluster,  # noqa: F401
    test_to_from_dict,  # noqa: F401
)


@pytest.fixture(autouse=True, scope="module")
def longer_timeout():
    # slurm tests started failing with timeouts
    # when adding timeout to test_restart_engines
    # maybe it's just slow...
    with mock.patch.object(test_cluster, "_timeout", 120):
        yield


# put ipython dir on shared filesystem
@pytest.fixture(autouse=True, scope="module")
def ipython_dir(request):
    if shutil.which("sbatch") is None:
        pytest.skip("Requires slurm")
    with temporary_ipython_dir(prefix="/data/") as ipython_dir:
        yield ipython_dir


@pytest.fixture
def cluster_config():
    c = Config()
    c.Cluster.controller_ip = '0.0.0.0'
    return c


# override launcher classes
@pytest.fixture
def engine_launcher_class():
    if shutil.which("sbatch") is None:
        pytest.skip("Requires slurm")
    return 'slurm'


@pytest.fixture
def controller_launcher_class():
    if shutil.which("sbatch") is None:
        pytest.skip("Requires slurm")
    return 'slurm'



import datetime
import time
from typing import Callable


def wait_for(condition: Callable):
    for _ in range(750):
        if condition():
            break
        else:
            time.sleep(0.1)
    if not condition():
        raise TimeoutError('wait_for took to long to finish')


def echo(delay=0):
    def inner_echo(x, **kwargs):
        import time

        if delay:
            time.sleep(delay)
        return x

    return inner_echo


def get_time_stamp() -> str:
    return (
        str(datetime.datetime.now()).split(".")[0].replace(" ", "-").replace(":", "-")
    )



import time
from subprocess import check_call

import ipyparallel as ipp

n = 20


check_call(f'ipcluster start -n {n} --daemon --profile=asv --debug', shell=True)
c = ipp.Client(profile='asv')
seen = -1

running_engines = len(c)

while running_engines < n:
    if seen != running_engines:
        print(running_engines)
        seen = running_engines
    running_engines = len(c)
    time.sleep(0.1)

check_call('ipcluster stop --profile=asv', shell=True)



import ipyparallel as ipp
from ipyparallel import Dependency

client = ipp.Client()


# this will only run on machines that can import numpy:
@ipp.require('numpy')
def norm(A):
    from numpy.linalg import norm

    return norm(A, 2)


def checkpid(pid):
    """return the pid of the engine"""
    import os

    return os.getpid() == pid


def checkhostname(host):
    import socket

    return socket.gethostname() == host


def getpid():
    import os

    return os.getpid()


pid0 = client[0].apply_sync(getpid)


# this will depend on the pid being that of target 0:
@ipp.depend(checkpid, pid0)
def getpid2():
    import os

    return os.getpid()


view = client.load_balanced_view()
view.block = True

# will run on anything:
pids1 = [view.apply(getpid) for i in range(len(client.ids))]
print(pids1)
# will only run on e0:
pids2 = [view.apply(getpid2) for i in range(len(client.ids))]
print(pids2)

print("now test some dependency behaviors")


def wait(t):
    import time

    time.sleep(t)
    return t


# fail after some time:
def wait_and_fail(t):
    import time

    time.sleep(t)
    return 1 / 0


successes = [view.apply_async(wait, 1).msg_ids[0] for i in range(len(client.ids))]
failures = [
    view.apply_async(wait_and_fail, 1).msg_ids[0] for i in range(len(client.ids))
]

mixed = [failures[0], successes[0]]
d1a = Dependency(mixed, all=False, failure=True)  # yes
d1b = Dependency(mixed, all=False)  # yes
d2a = Dependency(mixed, all=True, failure=True)  # yes after / no follow
d2b = Dependency(mixed, all=True)  # no
d3 = Dependency(failures, all=False)  # no
d4 = Dependency(failures, all=False, failure=True)  # yes
d5 = Dependency(failures, all=True, failure=True)  # yes after / no follow
d6 = Dependency(successes, all=True, failure=True)  # yes after / no follow

view.block = False
flags = view.temp_flags
with flags(after=d1a):
    r1a = view.apply(getpid)
with flags(follow=d1b):
    r1b = view.apply(getpid)
with flags(after=d2b, follow=d2a):
    r2a = view.apply(getpid)
with flags(after=d2a, follow=d2b):
    r2b = view.apply(getpid)
with flags(after=d3):
    r3 = view.apply(getpid)
with flags(after=d4):
    r4a = view.apply(getpid)
with flags(follow=d4):
    r4b = view.apply(getpid)
with flags(after=d3, follow=d4):
    r4c = view.apply(getpid)
with flags(after=d5):
    r5 = view.apply(getpid)
with flags(follow=d5, after=d3):
    r5b = view.apply(getpid)
with flags(follow=d6):
    r6 = view.apply(getpid)
with flags(after=d6, follow=d2b):
    r6b = view.apply(getpid)


def should_fail(f):
    try:
        f()
    except ipp.error.KernelError:
        pass
    else:
        print('should have raised')
        # raise Exception("should have raised")


# print(r1a.msg_ids)
r1a.get()
# print(r1b.msg_ids)
r1b.get()
# print(r2a.msg_ids)
should_fail(r2a.get)
# print(r2b.msg_ids)
should_fail(r2b.get)
# print(r3.msg_ids)
should_fail(r3.get)
# print(r4a.msg_ids)
r4a.get()
# print(r4b.msg_ids)
r4b.get()
# print(r4c.msg_ids)
should_fail(r4c.get)
# print(r5.msg_ids)
r5.get()
# print(r5b.msg_ids)
should_fail(r5b.get)
# print(r6.msg_ids)
should_fail(r6.get)  # assuming > 1 engine
# print(r6b.msg_ids)
should_fail(r6b.get)
print('done')



#!/usr/bin/env python3
import multiprocessing as mp
import os
import sys
from subprocess import Popen, check_call
from time import sleep
from typing import List

import googleapiclient.discovery as gcd

from benchmarks.utils import get_time_stamp

CORE_NUMBERS_FOR_TEMPLATES = [64]

ZONE = "europe-west1-b"
PROJECT_NAME = "jupyter-simula"
INSTANCE_NAME_PREFIX = "asv-testing-"
MACHINE_CONFIGS_DIR = os.path.join(os.getcwd(), "machine_configs")
CONDA_PATH = 'miniconda3/bin/conda'

compute = gcd.build("compute", "v1")


def generate_template_name(number_of_cores_and_ram):
    return f"{INSTANCE_NAME_PREFIX}{number_of_cores_and_ram}"


def get_running_instance_names() -> List[str]:
    result = compute.instances().list(project=PROJECT_NAME, zone=ZONE).execute()
    return [item["name"] for item in result["items"]] if "items" in result else []


def delete_instance(instance_name) -> dict:
    print(f"Deleting instance: {instance_name}")
    return (
        compute.instances()
        .delete(project=PROJECT_NAME, zone=ZONE, instance=instance_name)
        .execute()
    )


def delete_all_instances():
    return [
        delete_instance(name)
        for name in get_running_instance_names()
        if INSTANCE_NAME_PREFIX in name
    ]


def gcloud_run(*args, block=True):
    cmd = ["gcloud", "compute"] + list(args)
    print(f'$ {" ".join(cmd)}')
    (
        check_call(
            cmd
            # stdout=open(get_gcloud_log_file_name(instance_name) + ".log", "a+"),
            # stderr=open(f"{get_gcloud_log_file_name(instance_name)}_error.out", "a+"),
        )
        if block
        else Popen(cmd)
    )


def copy_files_to_instance(instance_name, *file_names, directory="~"):
    for file_name in file_names:
        gcloud_run("scp", file_name, f"{instance_name}:{directory}", f"--zone={ZONE}")


def command_over_ssh(instance_name, *args, block=True):
    return gcloud_run("ssh", instance_name, f"--zone={ZONE}", "--", *args, block=block)


def run_on_instance(template_name):
    current_instance_name = f"{template_name}-{get_time_stamp()}"
    benchmark_name = sys.argv[1] if len(sys.argv) > 1 else ''

    print(f"Creating new instance with name: {current_instance_name}")

    gcloud_run(
        "instances",
        "create",
        current_instance_name,
        "--source-instance-template",
        template_name,
    )
    sleep(20)  # Waiting for ssh keys to propagate to instance
    command_over_ssh(current_instance_name, "sudo", "apt", "update")
    print("copying instance setup to instance")
    command_over_ssh(
        current_instance_name,
        'wget',
        '-q',
        'https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh',
    )
    print('installing miniconda')
    command_over_ssh(
        current_instance_name, 'bash', 'Miniconda3-latest-Linux-x86_64.sh', '-b'
    )
    command_over_ssh(
        current_instance_name,
        'echo',
        '"source  miniconda3/bin/activate"',
        '>>',
        '~/.bashrc',
    )
    command_over_ssh(current_instance_name, f'{CONDA_PATH}', 'init')
    print('installing asv and google api')
    command_over_ssh(
        current_instance_name,
        f'{CONDA_PATH}',
        'install',
        '-c',
        'conda-forge',
        'asv',
        'google-api-python-client',
        'google-auth-httplib2',
        'google-auth-oauthlib',
        'google-cloud-storage',
        'numpy',
        '-y',
    )

    copy_files_to_instance(current_instance_name, "instance_setup.py")

    for config_name in os.listdir(MACHINE_CONFIGS_DIR):
        if config_name == template_name + ".json":
            copy_files_to_instance(
                current_instance_name,
                os.path.join(MACHINE_CONFIGS_DIR, config_name),
                directory="~/.asv-machine.json",
            )
            break
    else:
        print(f"Found no valid machine config for template: {template_name}.")
        exit(1)
    print("starting instance setup")
    command_over_ssh(
        current_instance_name,
        "miniconda3/bin/python3",
        "instance_setup.py",
        current_instance_name,
        template_name,
        block=True,
    )

    command_over_ssh(
        current_instance_name,
        'cd',
        'ipyparallel_master_project',
        ';',
        "~/miniconda3/bin/python3",
        'asv_runner.py',
        benchmark_name,
        template_name,
        block=False,
    )


if __name__ == "__main__":
    running_instances = get_running_instance_names()
    number_of_running_instances = len(running_instances)
    # atexit.register(delete_all_instances)

    print(f"Currently there are {number_of_running_instances} running instances.")
    if number_of_running_instances:
        print("Running instances: ")
        for instance in running_instances:
            print(f"  {instance}")

    if "-d" in sys.argv:
        result = delete_instance(sys.argv[2])
    elif "-da" in sys.argv:
        result = delete_all_instances()
    if "-q" in sys.argv:
        exit(0)

    if len(CORE_NUMBERS_FOR_TEMPLATES) == 1:
        run_on_instance(generate_template_name(CORE_NUMBERS_FOR_TEMPLATES[0]))
    else:
        with mp.Pool(len(CORE_NUMBERS_FOR_TEMPLATES)) as pool:
            result = pool.map_async(
                run_on_instance,
                [
                    generate_template_name(core_number)
                    for core_number in CORE_NUMBERS_FOR_TEMPLATES
                ],
            )
        result.wait()
    print("gcloud setup finished.")



"""Calculate statistics on the digits of pi in parallel.

This program uses the functions in :file:`pidigits.py` to calculate
the frequencies of 2 digit sequences in the digits of pi. The
results are plotted using matplotlib.

To run, text files from https://www.super-computing.org/
must be installed in the working directory of the IPython engines.
The actual filenames to be used can be set with the ``filestring``
variable below.

The dataset we have been using for this is the 200 million digit one here:
ftp://pi.super-computing.org/.2/pi200m/

and the files used will be downloaded if they are not in the working directory
of the IPython engines.
"""

from timeit import default_timer as clock

from matplotlib import pyplot as plt
from pidigits import (
    compute_two_digit_freqs,
    fetch_pi_file,
    plot_two_digit_freqs,
    reduce_freqs,
)

import ipyparallel as ipp

# Files with digits of pi (10m digits each)
filestring = 'pi200m.ascii.%(i)02dof20'
files = [filestring % {'i': i} for i in range(1, 21)]

# Connect to the IPython cluster
c = ipp.Client()
c[:].run('pidigits.py')

# the number of engines
n = len(c)
id0 = c.ids[0]
v = c[:]
v.block = True
# fetch the pi-files
print(f"downloading {n} files of pi")
v.map(fetch_pi_file, files[:n])  # noqa: F821
print("done")

# Run 10m digits on 1 engine
t1 = clock()
freqs10m = c[id0].apply_sync(compute_two_digit_freqs, files[0])
t2 = clock()
digits_per_second1 = 10.0e6 / (t2 - t1)
print("Digits per second (1 core, 10m digits):   ", digits_per_second1)


# Run n*10m digits on all engines
t1 = clock()
freqs_all = v.map(compute_two_digit_freqs, files[:n])
freqs150m = reduce_freqs(freqs_all)
t2 = clock()
digits_per_second8 = n * 10.0e6 / (t2 - t1)
print(f"Digits per second ({n} engines, {n}0m digits): ", digits_per_second8)

print("Speedup: ", digits_per_second8 / digits_per_second1)

plot_two_digit_freqs(freqs150m)
plt.title(f"2 digit sequences in {n}0m digits of pi")
plt.show()



import atexit
import os
import resource
import socket
import sys
from subprocess import check_call

import googleapiclient.discovery as gcd
from cluster_start import start_cluster
from google.cloud import storage

os.environ['OPENBLAS_NUM_THREADS'] = '1'

DEFAULT_MINICONDA_PATH = os.path.abspath(os.path.join('..', "miniconda3/bin/:"))
env = os.environ.copy()
env["PATH"] = DEFAULT_MINICONDA_PATH + env["PATH"]

ZONE = "europe-west1-b"
PROJECT_NAME = "jupyter-simula"
BUCKET_NAME = 'ipyparallel_dev'

instance_name = socket.gethostname()
compute = gcd.build("compute", "v1")


def cmd_run(*args, log_filename=None, error_filename=None):
    if len(args) == 1:
        args = args[0].split(" ")
    print(f'$ {" ".join(args)}')
    if not log_filename and not error_filename:
        check_call(args, env=env)
    else:
        check_call(
            args,
            env=env,
            stdout=open(log_filename, 'w'),
            stderr=open(error_filename, 'w'),
        )


def delete_self():
    print(f'deleting: {instance_name}')
    compute.instances().delete(
        project=PROJECT_NAME, zone=ZONE, instance=instance_name
    ).execute()


def upload_file(filename):
    blob_name = sys.argv[1]
    storage_client = storage.Client()
    bucket = storage_client.bucket(BUCKET_NAME)
    blob = bucket.blob(f'{instance_name}/{blob_name}')
    print(f'Uploading {filename}')
    blob.upload_from_filename(filename)


if __name__ == '__main__':
    # atexit.register(delete_self)
    benchmark_name = sys.argv[1]
    template_name = sys.argv[2]

    soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
    resource.setrlimit(resource.RLIMIT_NOFILE, (hard_limit, hard_limit))

    ps = []
    for i in [3, 4]:
        ps += start_cluster(i, 1030, '../miniconda3/bin/', log_output_to_file=True)
    # time.sleep(10)
    # ps += start_cluster(
    #     0, 'depth_0', 300, '../miniconda3/bin/',
    #     log_output_to_file=True)
    # time.sleep(10)
    # ps += start_cluster(
    #     0, 'depth_1', 150, '../miniconda3/bin/',
    #     log_output_to_file=True)

    log_filename = f'{instance_name}.log'
    error_log_filename = f'{instance_name}.error.log'

    def clean_up():
        for p in ps:
            p.kill()

    atexit.register(clean_up)
    # cmd_run("ipcluster start -n 200 --daemon --profile=asv")  # Starting 200 engines
    cmd_run(
        'asv run -b DepthTestingSuite --verbose --show-stderr',
        # log_filename=log_filename,
        # error_filename=error_log_filename,
    )
    clean_up()
    # cmd_run("ipcluster stop --profile=asv")
    print('uploading files')
    # upload_file(log_filename)
    # upload_file(error_log_filename)
    results_dir = f'results/{template_name}'
    for file_name in os.listdir(results_dir):
        if 'conda' in file_name:
            upload_file(f'{results_dir}/{file_name}')
    print('script finished')



#!/usr/bin/env python
"""
Receiver side of OS independent shell commands

For a list of supported commands see top of file shellcmd.py.

Important: The shellcmd concept also supports receiver code sending (useful for testing and developing)
which transfers this file to the 'other side'. However, this limits the imports to standard python
packages. Hence, DO NOT USE ANY ipyparallel IMPORTS IN THIS FILE!
"""

import logging
import os
import shutil
import sys
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from random import randint
from subprocess import DEVNULL, Popen


class ShellCommandReceiveBase(metaclass=ABCMeta):
    """
    Base class for receiving and performing shell commands in a platform independent form

    All supported shell commands have a cmd_ prefix. When adding new functions make sure that there is an
    equivalent in the ShellCommandSend class. When a command failed a non-zero exit code will be returned.
    Hence, the ShellCommandSend class always uses subprocess.check_output for assessing if the command was
    successful. Some command require information to be returned (cmd_exists, cmd_start, cmd_running) which
    is written to stdout in the following form: __<key>=<value>__

    This base class contains all platform independent code and commands. Only, cmd_start, cmd_running and
    cmd_kill needs to be overwritten in childern classes (see ShellCommandReceiveWindows and
    ShellCommandReceivePosix)
    """

    def _log(self, msg):
        if not self.log:
            return
        self.log.info(f"[id={self.ranid}] {msg}")

    def __init__(self, debugging=False, log=None):
        self.debugging = debugging
        self.log = None
        if log is None:
            log = os.getenv("SHELLCMD_LOG")
        if log:
            log_file = Path(log).expanduser()
            self.log = logging.getLogger("shellcmd")
            self.log.setLevel(logging.DEBUG if self.debugging else logging.INFO)
            self.log.addHandler(logging.FileHandler(log_file, mode="a"))

        self.ranid = None
        if self.log:
            self.ranid = randint(0, 999)
            self._log(f"{self.__class__.__name__} instance created")

    def close(self):
        # perform possible clean up actions (currently not required)
        if self.log:
            self._log("ShellCommandReceiveBase closed")

    def _prepare_cmd_start(self, start_cmd, env):
        if env:
            self._log(f"env={env!r}")
            if not isinstance(env, dict):
                raise TypeError(f"env must be a dict, got {env!r}")

            # update environment
            for key, value in env.items():
                if value is not None and value != '':
                    # set entry
                    os.environ[key] = str(value)
                else:
                    # unset entry if needed
                    if key in os.environ:
                        del os.environ[key]

        if isinstance(start_cmd, str):
            start_cmd = [start_cmd]

        if not all(isinstance(item, str) for item in start_cmd):
            raise TypeError(f"Only str in start_cmd allowed ({start_cmd!r})")

        return start_cmd

    @abstractmethod
    def cmd_start(self, cmd, env=None, output_file=None):
        pass

    @abstractmethod
    def cmd_running(self, pid):
        pass

    @abstractmethod
    def cmd_kill(self, pid, sig=None):
        pass

    def cmd_mkdir(self, path):
        self._log(f"Make directory '{path}'")
        os.makedirs(path, exist_ok=True)  # we allow that the directory already exists

    def cmd_rmdir(self, path):
        self._log(f"Remove directory '{path}'")
        shutil.rmtree(path)

    def cmd_exists(self, path):
        self._log(f"Check if path exists '{path}'")
        if os.path.exists(path):
            print("__exists=1__")
        else:
            print("__exists=0__")

    def cmd_remove(self, path):
        self._log(f"Remove file '{path}'")
        os.remove(path)


class ShellCommandReceiveWindows(ShellCommandReceiveBase):
    """Windows Implementation of ShellCommandReceive class"""

    def __init__(self, debugging=False, use_breakaway=True, log=None):
        super().__init__(debugging, log)
        self.use_breakaway = use_breakaway

    def cmd_start(self, cmd, env=None, output_file=None):
        start_cmd = self._prepare_cmd_start(cmd, env)

        # under windows we need to remove embracing double quotes
        for idx, p in enumerate(start_cmd):
            if p[0] == '"' and p[-1] == '"':
                start_cmd[idx] = p.strip('"')

        self._log(f"start_cmd={start_cmd}  (use_breakaway={self.use_breakaway})")

        from subprocess import CREATE_BREAKAWAY_FROM_JOB, CREATE_NEW_CONSOLE

        flags = 0
        if self.use_breakaway:
            flags |= CREATE_NEW_CONSOLE
            flags |= CREATE_BREAKAWAY_FROM_JOB

        pkwargs = {
            'close_fds': True,  # close stdin/stdout/stderr on child
            'creationflags': flags,
        }
        if output_file:
            fo = open(output_file, "w")
            pkwargs['stdout'] = fo
            pkwargs['stderr'] = fo
            pkwargs['stdin'] = DEVNULL

        self._log(f"Popen(**pkwargs={pkwargs})")
        p = Popen(start_cmd, **pkwargs)
        self._log(f"pid={p.pid}")

        print(f'__remote_pid={p.pid}__')
        sys.stdout.flush()
        if not self.use_breakaway:
            self._log("before wait")
            p.wait()
            self._log("after wait")

    def cmd_running(self, pid):
        self._log(f"Check if pid {pid} is running")

        # taken from https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python
        import ctypes

        PROCESS_QUERY_INFROMATION = (
            0x1000  # if actually PROCESS_QUERY_LIMITED_INFORMATION
        )
        STILL_ACTIVE = 259
        processHandle = ctypes.windll.kernel32.OpenProcess(
            PROCESS_QUERY_INFROMATION, 0, pid
        )
        if processHandle == 0:
            print('__running=0__')
        else:
            i = ctypes.c_int(0)
            pi = ctypes.pointer(i)
            if ctypes.windll.kernel32.GetExitCodeProcess(processHandle, pi) == 0:
                print('__running=0__')
            if i.value == STILL_ACTIVE:
                print('__running=1__')
            else:
                print('__running=0__')
            ctypes.windll.kernel32.CloseHandle(processHandle)

    def cmd_kill(self, pid, sig=None):
        self._log(f"Kill pid {pid} (signal={sig})")

        # os.kill doesn't work reliable under windows. also see
        # https://stackoverflow.com/questions/28551180/how-to-kill-subprocess-python-in-windows

        # solution using taskill
        # import subprocess
        # subprocess.call(['taskkill', '/F', '/T', '/PID',  str(pid)])  # /T kills all child processes as well

        # use windows api to kill process (doesn't kill children processes)
        # To kill all children process things are more complicated. see e.g.
        # http://mackeblog.blogspot.com/2012/05/killing-subprocesses-on-windows-in.html
        import ctypes

        PROCESS_TERMINATE = 0x0001
        kernel32 = ctypes.windll.kernel32
        processHandle = kernel32.OpenProcess(PROCESS_TERMINATE, 0, pid)
        if processHandle:
            kernel32.TerminateProcess(
                processHandle, 3
            )  # 3 is just an arbitrary exit code
            kernel32.CloseHandle(processHandle)


class ShellCommandReceivePosix(ShellCommandReceiveBase):
    """Posix implementation of the ShellCommandReceive class"""

    def cmd_start(self, cmd, env=None, output_file=None):
        start_cmd = self._prepare_cmd_start(cmd, env)

        fo = DEVNULL
        if output_file:
            fo = open(output_file, "w")

        p = Popen(
            start_cmd, start_new_session=True, stdout=fo, stderr=fo, stdin=DEVNULL
        )
        print(f'__remote_pid={p.pid}__')
        sys.stdout.flush()

    def cmd_running(self, pid):
        self._log(f"Check if pid {pid} is running")
        try:
            # use os.kill with signal 0 to check if process is still running
            os.kill(pid, 0)
            print('__running=1__')
        except OSError:
            print('__running=0__')

    def cmd_kill(self, pid, sig=None):
        self._log(f"Kill pid {pid} (signal={sig})")
        os.kill(pid, sig)


@contextmanager
def ShellCommandReceive(debugging=False, use_breakaway=True, log=None):
    """Generator returning the corresponding platform dependent ShellCommandReceive object (as Context Manager)"""
    if sys.platform.lower().startswith("win"):
        receiver = ShellCommandReceiveWindows(debugging, use_breakaway, log)
    else:
        receiver = ShellCommandReceivePosix(debugging, log)
    try:
        yield receiver
    finally:
        receiver.close()



#!/usr/bin/env python
"""Parallel word frequency counter.

This only works for a local cluster, because the filenames are local paths.
"""

import os
import time
from itertools import repeat

import requests
from wordfreq import print_wordfreq, wordfreq

import ipyparallel as ipp

davinci_url = "https://www.gutenberg.org/files/5000/5000-8.txt"


def pwordfreq(view, fnames):
    """Parallel word frequency counter.

    view - An IPython DirectView
    fnames - The filenames containing the split data.
    """
    assert len(fnames) == len(view.targets)
    view.scatter('fname', fnames, flatten=True)
    ar = view.apply(wordfreq, ipp.Reference('fname'))
    freqs_list = ar.get()
    word_set = set()
    for f in freqs_list:
        word_set.update(f.keys())
    freqs = dict(zip(word_set, repeat(0)))
    for f in freqs_list:
        for word, count in f.items():
            freqs[word] += count
    return freqs


if __name__ == '__main__':
    # Create a Client and View
    rc = ipp.Client()

    view = rc[:]
    view.apply_sync(os.chdir, os.getcwd())

    if not os.path.exists('davinci.txt'):
        # download from project gutenberg
        print("Downloading Da Vinci's notebooks from Project Gutenberg")
        r = requests.get(davinci_url)
        with open('davinci.txt', 'w', encoding='utf8') as f:
            f.write(r.text)

    # Run the serial version
    print("Serial word frequency count:")
    text = open('davinci.txt', encoding='latin1').read()
    tic = time.time()
    freqs = wordfreq(text)
    toc = time.time()
    print_wordfreq(freqs, 10)
    print(f"Took {toc - tic:.3f}s to calculate")

    # The parallel version
    print("\nParallel word frequency count:")
    # split the davinci.txt into one file per engine:
    lines = text.splitlines()
    nlines = len(lines)
    n = len(rc)
    block = nlines // n
    for i in range(n):
        chunk = lines[i * block : i * (block + 1)]
        with open(f'davinci{i}.txt', 'w', encoding='utf8') as f:
            f.write('\n'.join(chunk))

    try:  # python2
        cwd = os.path.abspath(os.getcwdu())
    except AttributeError:  # python3
        cwd = os.path.abspath(os.getcwd())
    fnames = [os.path.join(cwd, f'davinci{i}.txt') for i in range(n)]
    tic = time.time()
    pfreqs = pwordfreq(view, fnames)
    toc = time.time()
    print_wordfreq(freqs)
    print(f"Took {toc - tic:.3f} s to calculate on {len(view.targets)} engines")
    # cleanup split files
    map(os.remove, fnames)



"""An example for handling results in a way that AsyncMapResult doesn't provide

Specifically, out-of-order results with some special handing of metadata.

This just submits a bunch of jobs, waits on the results, and prints the stdout
and results of each as they finish.

Authors
-------
* MinRK
"""

import random

import ipyparallel as ipp

# create client & views
rc = ipp.Client()
dv = rc[:]
v = rc.load_balanced_view()


# scatter 'id', so id=0,1,2 on engines 0,1,2
dv.scatter('id', rc.ids, flatten=True)
print(dv['id'])


def sleep_here(count, t):
    """simple function that takes args, prints a short message, sleeps for a time, and returns the same args"""
    import sys
    import time

    print(f"hi from engine {id}")
    sys.stdout.flush()
    time.sleep(t)
    return count, t


amr = v.map(sleep_here, range(100), [random.random() for i in range(100)], chunksize=2)

pending = set(amr.msg_ids)
while pending:
    try:
        rc.wait(pending, 1e-3)
    except TimeoutError:
        # ignore timeouterrors, since they only mean that at least one isn't done
        pass
    # finished is the set of msg_ids that are complete
    finished = pending.difference(rc.outstanding)
    # update pending to exclude those that just finished
    pending = pending.difference(finished)
    for msg_id in finished:
        # we know these are done, so don't worry about blocking
        ar = rc.get_result(msg_id)
        print(f"job id {msg_id} finished on engine {ar.engine_id}")
        print("with stdout:")
        print('    ' + ar.stdout.replace('\n', '\n    ').rstrip())
        print("and results:")

        # note that each job in a map always returns a list of length chunksize
        # even if chunksize == 1
        for count, t in ar.get():
            print(f"  item {count}: slept for {t:.2f}s")



import datetime
import os
import sys
import time
from subprocess import Popen, check_call, check_output

import ipyparallel as ipp


def wait_for(condition):
    for _ in range(750):
        if condition():
            break
        else:
            time.sleep(0.1)
    if not condition():
        raise TimeoutError('wait_for took to long to finish')


def get_time_stamp() -> str:
    return (
        str(datetime.datetime.now()).split(".")[0].replace(" ", "-").replace(":", "-")
    )


def start_cmd(cmd, blocking=True):
    print(cmd)
    return (
        check_call(
            cmd,
            stdout=sys.__stdout__,
            stderr=open('spanning_tree_error.out', 'a+'),
            shell=True,
        )
        if blocking
        else Popen(
            cmd,
            stdout=sys.__stdout__,
            stderr=open('spanning_tree_error.out', 'a+'),
            shell=True,
        )
    )


def stop_cluster():
    if '-s' not in sys.argv:
        start_cmd('ipcluster stop --profile=asv')


# atexit.register(stop_cluster)

PROFILING_TASKS = [
    'many_empty_tasks',
    'many_empty_tasks_non_blocking',
    'tasks_with_large_data',
    'echo_many_arguments',
]

VIEW_TYPES = ['direct', 'load_balanced']


def get_tasks_to_execute(program_arguments):
    return (
        [f'{program_arguments[2]} {view}' for view in VIEW_TYPES]
        if len(program_arguments) >= 2
        else [f'{task} {view}' for task in PROFILING_TASKS for view in VIEW_TYPES]
    )


if __name__ == "__main__":
    print('profiling_code_runner_started')
    if '-s' not in sys.argv:
        n = int(sys.argv[1]) if len(sys.argv) > 1 else 16
        client = ipp.Client(profile='asv')
        print(f'Waiting for {n} engines to get available')
        try:
            wait_for(lambda: len(client) >= n)
        except TimeoutError as e:
            print(e)
            exit(1)
        print('Starting the profiling')

    controller_pid = check_output('pgrep -f ipyparallel.controller', shell=True)
    number_of_schedulers = 15
    scheduler_pids = sorted(int(x) for x in controller_pid.decode('utf-8').split())[
        -number_of_schedulers:
    ]

    client_output_path = os.path.join(os.getcwd(), 'spanning_tree_client.svg')

    files_to_upload = [client_output_path]
    ps = []
    for i, scheduler_pid in enumerate(scheduler_pids):
        scheduler_output_path = os.path.join(
            os.getcwd(), f'spanning_tree_{i}_scheduler.svg'
        )
        files_to_upload.append(scheduler_output_path)
        ps.append(
            start_cmd(
                f'sudo py-spy --function -d 60 --flame {scheduler_output_path} --pid {scheduler_pid}',
                blocking=False,
            )
        )

    start_cmd(
        f'sudo py-spy --function -d 60 --flame {client_output_path} -- python profiling_code.py tasks_with_large_data spanning_tree'
    )
    print('client ended')
    for p in ps:
        p.wait()
